/**
  *  \file client/dialogs/clonedlg.cc
  *  \brief Ship Cloning
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/dialogs/clonedlg.h"
#include "client/dialogs/baseshipbuild.h"
#include "client/dlg-tech.h"
#include "cpluslib/format.h"
#include "game/actions/basetech.h"
#include "game/actions/planetsupplies.h"
#include "game/global.h"
#include "game/host.h"
#include "game/mission.h"
#include "game/planet.h"
#include "game/reg.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/widgets/keytranslator.h"
#include "ui/widgets/multilinestatic.h"
#include "ui/window.h"
#include "util/nls.h"
#include "game/fleet.h"
#include "game/config.h"
#include "game/planettransfer.h"
#include "ui/widgets/quit.h"

namespace {
    class WCloneCargoCostTransaction : public GCargoCostTransaction {
        GPlanet& planet;
        GShip& ship;
        int32_t ccr;

        GCost shipCost;
        GStarbaseTechUpgradeAction techUpgrade;

        bool techFailure;
     public:
        WCloneCargoCostTransaction(GCargoContainer& container, GPlanet& planet, GShip& ship, int32_t ccr);
        void update();

        bool isTechFailure() const
            { return techFailure; }
        bool isTechUpgradable()
            { return getReservedCargo(el_Money) + getReservedCargo(el_Supplies) >= techUpgrade.getCost().get(el_Money); }
        bool isAnyTechUpgrade()
            { return techUpgrade.getCost().get(el_Money) > 0; }
        GStarbaseTechUpgradeAction& getTechUpgrade()
            { return techUpgrade; }
        const GCost& getShipCost() const
            { return shipCost; }
    };
}

WCloneCargoCostTransaction::WCloneCargoCostTransaction(GCargoContainer& container, GPlanet& planet, GShip& ship, int32_t ccr)
    : GCargoCostTransaction(container),
      planet(planet),
      ship(ship),
      ccr(ccr),
      shipCost(),
      techUpgrade(planet),
      techFailure(true)
{
    update();
}

void
WCloneCargoCostTransaction::update()
{
    // Compute ship cost
    int needTech[4];
    const GHull& hull = ship.getHull();

    shipCost = hull.getCost();
    needTech[tl_Hull] = hull.getTech();

    shipCost += getEngine(ship.getEngineType()).getCost() * hull.getNumEngines();
    needTech[tl_Engine] = getEngine(ship.getEngineType()).getTech();

    if (ship.getNumBeams() > 0) {
        shipCost += getBeam(ship.getBeamType()).getCost() * ship.getNumBeams();
        needTech[tl_Beam] = getBeam(ship.getBeamType()).getTech();
    } else {
        needTech[tl_Beam] = 1;
    }

    if (ship.getNumTorpLaunchers() > 0) {
        shipCost += getTorp(ship.getTorpType()).getCost() * ship.getNumTorpLaunchers();
        needTech[tl_Torpedo] = getTorp(ship.getTorpType()).getTech();
    } else {
        needTech[tl_Torpedo] = 1;
    }
    
    // Apply CloneCostRate. We must be careful to avoid overflow.
    // The naive computation, mc*ccr/100, would overflow when
    // ccr=32767 and mc>65536. The most expensive ship known to me
    // is a fully-loaded T10 Basilisk at $35k, so it's not hard to
    // invent a $70k ship. This computation multiplies with no more
    // than 327, allowing for ship costs up to $6.5M.
    shipCost.set(el_Money,
                 shipCost.get(el_Money) * (ccr / 100)
                 + (shipCost.get(el_Money) * (ccr % 100)) / 100);
    
    // Check tech
    techFailure = false;
    for (int i = 0; i < 4; ++i) {
        if (needTech[i] > techUpgrade.getMinTech(i)) {
            if (needTech[i] > techUpgrade.getMaxTech(i)) {
                techFailure = true;
            } else {
                techUpgrade.setTech(i, needTech[i]);
            }
        }
    }

    // Configure cargo cost
    GCost total(shipCost);
    total += techUpgrade.getCost();
    setCost(total);
}


/** Clone a ship. This decides how to clone the ship, and branches to the
    correct dialog.
    - ships that can be normally built will open the ship build dialog with
      the ship predefined
    - ships that can be cloned will be prepared for cloning
    - remaining ships will receive an error message (can happen with PHost
      remote control).
    \param sh Ship to clone
    \param pl Planet to work on */
void
doCloneShip(GShip& sh, GPlanet& pl)
{
    // Validate
    if (!sh.isPlayable(GObject::Playable) || !pl.isPlayable(GObject::Playable) || !pl.hasBase()) {
        return;
    }

    // First, check previous orders
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    const int previous = pl.findShipCloningHere(univ);
    const int thslot = getTruehullSlot(pl.getOwner(), sh.getHullId());
    if (previous != 0 && previous != sh.getId()) {
        // It's cloning
        if (!UIMessageBox(format(_("This base is already cloning %s. Proceed anyway and replace this order?"), univ.getShip(previous).getName(GObject::PlainName)),
                          _("Clone Ship")).doYesNoDialog())
        {
            return;
        }
    } else if (pl.getBaseBuildHull() != 0) {
        // It's building normally
        const char* msg;
        if (thslot == 0) {
            msg = _("This base is already building a %s. Proceed anyway and delay this order?");
        } else {
            msg = _("This base is already building a %s. Proceed anyway and replace this order?");
        }
        if (!UIMessageBox(format(msg, getHull(pl.getBaseBuildHull()).getName()), _("Clone Ship")).doYesNoDialog()) {
            return;
        }
    }

    // What do we do?
    if (thslot != 0) {
        // We want to build a copy of a ship. Prepare the build order
        TBuildOrder order;
        order.hull       = sh.getHullId();
        order.engine     = sh.getEngineType();
        order.beam_type  = sh.getBeamType();
        order.beam_count = sh.getNumBeams();
        order.torp_type  = sh.getTorpType();
        order.torp_count = sh.getNumTorpLaunchers();
        order.unused     = 0;

        // Prepare build dialog.
        std::vector<int> hulls;
        for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i) {
            if (int j = getTruehull(pl.getOwner(), i)) {
                hulls.push_back(j);
            }
        }
        WBaseShipBuildDialog bs(pl, hulls);
        bs.setBuildOrder(order);
        ui_root->centerWidget(bs);
        ui_root->add(bs);
        ui_root->run();
        ui_root->close(bs);
    } else {
        // Check registration status
        const int rso = sh.getRealOwner();
        if (!allowRegistered(rso)) {
            messageBox(_("You cannot clone ships."), _("Clone Ship"));
            return;
        }

        // Check clone cost rate
        int32_t ccr;
        if (host.isTHost()) {
            // Tim-Host: some players may not be able to clone
            if (raceId(rso) == 5 || raceId(rso) == 7) {
                messageBox(_("You cannot clone ships."), _("Clone Ship"));
                return;
            }
            ccr = 200;
        } else {
            // PHost: everyone can clone (at possibly prohibitive prices),
            if (getTruehullSlot(rso, sh.getHullId()) != 0) {
                messageBox(_("This ship cannot be cloned here, because its real owner "
                             "cannot clone it. You have to own it yourself to be able "
                             "to clone it."), _("Clone Ship"));
                return;
            }
            ccr = config.ShipCloneCostRate(rso);
        }

        // Check hull functions
        if (sh.canDoSpecial(hf_Unclonable)) {
            messageBox(_("This ship is unclonable."),
                       _("Clone Ship"));
            return;
        }

        // Payment logic. We build a specialized cargo cost transaction just
        // for the purpose of display. This transaction will never be committed.
        GMultiTransaction mt;
        GPlanetTransfer& tx =
            mt.addNewSubtransaction(new GPlanetTransfer(pl));
        WCloneCargoCostTransaction& ccct =
            mt.addNewSubtransaction(new WCloneCargoCostTransaction(tx, pl, sh, ccr));

        if (ccct.isTechFailure()) {
            mt.doCancel();
            messageBox(_("To clone this ship, you need to upgrade your technology over Tech 6. "
                         "As a shareware player, you can''t do this."),
                       _("Build order rejected"));
            return;
        }

        // Dialog
        //   WCostDisplay
        //   UIMultilineStatic
        //   HBox
        //     UISpacer
        //     "OK" or "Yes"/"No"
        //     UISpacer
        ObjectHolder h;
        UIWindow& win = h.add(new UIWindow(_("Clone Ship"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
        win.setLayout(UIVBoxLayout::instance5);

        WCostDisplay& costdpy = h.add(new WCostDisplay(ccct, GCargoTypeSet() + el_Tritanium + el_Duranium + el_Molybdenum + el_Supplies + el_Money));
        win.add(costdpy);
        bool ok;
        string_t message;
        if (!ccct.isTechUpgradable()) {
            // Not enough cash to upgrate tech
            ok = false;
            message = _("You cannot clone this ship, because you do not have enough money to upgrade your technology.");
        } else if (!ccct.isValidCost()) {
            // Can upgrade tech, but not build ship
            ok = true;
            message = _("You do not have enough resources to clone this ship. Try anyway?");
            if (ccct.isAnyTechUpgrade()) {
                message += _(" At this point, PCC2 will buy only the required tech upgrades.");
            }
        } else {
            // Everything fine
            ok = true;
            message = _("Build this ship?");
        }
        
        UIWidget::LayoutInfo info;
        costdpy.getLayoutInfo(info);
        win.add(h.add(new UIMultilineStatic(FONT_NORMAL_BOLD, message, info.pref_size.x)));
        if (ok && sh.canDoSpecial(hf_CloneOnce)) {
            win.add(h.add(new UIMultilineStatic(FONT_NORMAL, _("Note: this ship can be cloned only once; it will be unclonable after the clone."), info.pref_size.x)));
        }
        UIGroup& g3 = h.add(new UIGroup(UIHBoxLayout::instance5));
        g3.add(h.add(new UISpacer()));
        if (ok) {
            g3.add(h.add(new UIButton(1, UIKeyString("Yes"))).addStop());
            g3.add(h.add(new UIButton(0, UIKeyString("No"))).addStop());
        } else {
            g3.add(h.add(new UIButton(0, UIKeyString("OK"))).addStop());
        }
        g3.add(h.add(new UISpacer()));
        win.add(g3);
        win.add(h.add(new UIKeyTranslator())
                .addStop(SDLK_RETURN, 1)
                .addStop(' ', 1)
                .addStop(SDLK_ESCAPE, 0));
        win.add(h.add(new UIQuit(0)));
        win.pack();
        ui_root->centerWidget(win);

        // Display dialog.
        bool result = ui_root->runModal(win);
        mt.doCancel();
        if (!result || !ok) {
            return;
        }

        // Do it. The original transaction has been cancelled, but we can reuse
        // the GStarbaseTechUpgradeAction to do the tech upgrades.
        GMultiTransaction mt2;
        GPlanetTransfer& tx2 =
            mt2.addNewSubtransaction(new GPlanetTransfer(pl));
        mt2.addNewSubtransaction(new GStarbaseBuildTransaction(ccct.getTechUpgrade(), tx2));
        if (GError e = mt2.getError()) {
            // Can't happen?
            mt2.doCancel();
            messageBox(e.getUserError(), _("Build order rejected"));
            return;
        }
        mt2.doCommit();

        // Sell supplies to have enough cash available.
        int32_t cashNeeded = ccct.getShipCost().get(el_Money);
        int32_t cashAvail = pl.getCargo(el_Money);
        if (cashAvail < cashNeeded) {
            GPlanetSellSuppliesAction pssa(pl);
            pssa.setAmount(std::min(pssa.getMaximumAmount(), cashNeeded - cashAvail));
            if (GError e = pssa.getError()) {
                // Can't happen?
                pssa.doCancel();
                messageBox(e.getUserError(), _("Build order rejected"));
                return;
            }
            pssa.doCommit();
        }

        // Check fleet
        if (sh.getFleetNumber() != 0) {
            if (!UIMessageBox(_("This ship is member of a fleet. To clone, it must leave the fleet and stay here. "
                                "Leave the fleet?"),
                              _("Clone Ship")).doYesNoDialog())
            {
                return;
            }
            leaveFleet(univ, sh.getId());
        }

        // Give the order
        cancelAllCloneOrders(pl);
        sh.setFCode("cln");
        setFleetWaypoint(univ, sh.getId(), sh.getPos());
        setFleetSpeed(univ, sh.getId(), 0);
        if (sh.getMission() == GMission::msn_Intercept) {
            // FIXME: Auto-task (mutex) awareness! (but this is missing in 1.x as well)
            setFleetMission(univ, sh.getId(), 0, 0, 0);
        }
    }
}
