/**
  *  \file client/dlg-tax.cc
  *  \brief Taxation Dialog
  *
  *  This contains widgets and a complete dialog to edit taxes on a
  *  planet.
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-tax.h"
#include "cpluslib/format.h"
#include "game/host.h"
#include "game/global.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/skin.h"
#include "ui/util.h"
#include "util/nls.h"
#include "game/tables.h"
#include "ui/prefix.h"
#include "game/racename.h"
#include "ui/widgets/focusgroup.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"
#include "client/tiles/planetgrowth.h"
#include "client/widgets/helpwidget.h"
#include "game/planetform.h"
#include "util/misc.h"
#include "game/mission.h"
#include "ui/widgets/listlikenumbersel.h"
#include "ui/widgets/quit.h"

namespace {
    const int NUM_EFF = 4;

    GHull* findSpecimenHull(int basicFunctionId)
    {
        GHull* result = 0;
        for (int i = 1, n = getNumHulls(); i <= n; ++i) {
            GHull& candidate = getHull(i);
            GPlayerSet set = candidate.getPlayersThatCan(basicFunctionId, true, GExpLevelSet(0));
            set += 0;
            if (set.contains(GPlayerSet::allUpTo(NUM_PLAYERS))) {
                if (result == 0) {
                    // First candidate
                    result = &candidate;
                } else {
                    // Ambiguous
                    result = 0;
                    break;
                }
            }
        }
        return result;
    }

    string_t addHullName(string_t text, int basicFunctionId)
    {
        if (GHull* h = findSpecimenHull(basicFunctionId)) {
            text += format(" (%s)", h->getShortName());
        }
        return text;
    }

    class WEffectorDialog : public UIWindow {
     public:
        WEffectorDialog(GPlanetEffectors& eff, Signal1<GPlanetEffectors>& sig)
            : UIWindow(_("Prediction Effects"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
              h(), eff(eff), sig(sig)
            {
                init();
            }

        void init()
            {
                editors[0] = &h.add(new UIListLikeNumberSelector(1, _("Ships hissing"), 0, 100, 1));
                editors[1] = &h.add(new UIListLikeNumberSelector(2, addHullName(_("Ships cooling to 50\xC2\xB0\x46"), hf_CoolsTo50),   0, 100, 1));
                editors[2] = &h.add(new UIListLikeNumberSelector(3, addHullName(_("Ships heating to 50\xC2\xB0\x46"), hf_HeatsTo50),   0, 100, 1));
                editors[3] = &h.add(new UIListLikeNumberSelector(4, addHullName(_("Ships heating to 100\xC2\xB0\x46"), hf_HeatsTo100), 0, 100, 1));

                setLayout(UIVBoxLayout::instance5);

                UIGroup& group_edit    = h.add(new UIGroup(UIVBoxLayout::instance0));
                UIGroup& group_buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
                for (int i = 0; i < NUM_EFF; ++i) {
                    group_edit.add(*editors[i]);
                    editors[i]->sig_changed.add(this, &WEffectorDialog::updateTo);
                }
                group_edit.add(h.add(new UIFocusIterator(1, NUM_EFF, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Page | UIFocusIterator::fi_Home)));
                group_buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
                group_buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
                group_buttons.add(h.add(new UISpacer()));
                add(group_edit);
                add(group_buttons);
                add(h.add(new UIQuit(cm_Escape)));

                updateFrom();

                int focus = 0;
                for (int i = 1; i < NUM_EFF; ++i) {
                    if (editors[i]->getValue() != 0) {
                        focus = i;
                        break;
                    }
                }
                editors[focus]->setFocusedWidget();
            }

        void updateFrom()
            {
                editors[0]->setValue(eff.get(GPlanetEffectors::pe_Hiss));
                editors[1]->setValue(eff.get(GPlanetEffectors::pe_Cool50));
                editors[2]->setValue(eff.get(GPlanetEffectors::pe_Heat50));
                editors[3]->setValue(eff.get(GPlanetEffectors::pe_Heat100));
            }

        void updateTo()
            {
                eff.set(GPlanetEffectors::pe_Hiss,    editors[0]->getValue());
                eff.set(GPlanetEffectors::pe_Cool50,  editors[1]->getValue());
                eff.set(GPlanetEffectors::pe_Heat50,  editors[2]->getValue());
                eff.set(GPlanetEffectors::pe_Heat100, editors[3]->getValue());
                sig.raise(eff);
            }

     private:
        ObjectHolder h;
        GPlanetEffectors& eff;
        Signal1<GPlanetEffectors>& sig;
        UIListLikeNumberSelector* editors[NUM_EFF];
    };
}


static int
getHissEffect(const GPlanet& planet, const GPlanetEffectors& eff)
{
    if (config.AllowHiss()) {
        int nhiss = eff.get(GPlanetEffectors::pe_Hiss);
        if (host.isPHost() && nhiss > config.MaxShipsHissing())
            nhiss = config.MaxShipsHissing();
        return nhiss * config.HissEffectRate(planet.getOwner());
    } else {
        return 0;
    }
}

/** Display change in happiness.
    \param ctx    context
    \param x,y,w  location and size
    \param change numeric happiness change */
static void
showChange(GfxContext& ctx, int x, int y, int w, int change)
{
    ctx.setColor(change < 0 ? UIColor::tc_Red : UIColor::tc_Green);
    if (!change)
        outTextF(ctx, x, y, w, getHappyChangeName(change));
    else
        outTextF(ctx, x, y, w, format("%s (%d)", getHappyChangeName(change), change));
}

static void
changeColonistTax(UINumberSelector& sel, const GPlanet& planet, int delta)
{
    int tax = sel.getValue();
    int32 limit;
    int32 due = getColonistDueLimited(planet, tax, limit);
    do
        tax += delta;
    while (tax > 0 && tax < 100 && getColonistDueLimited(planet, tax, limit) == due);
    sel.setValue(tax);
}

static void
changeNativeTax(UINumberSelector& sel, int ctax, const GPlanet& planet, int delta)
{
    int tax = sel.getValue();
    int32 limit;
    getColonistDueLimited(planet, ctax, limit);
    int due = getNativeDueLimited(planet, tax, limit);
    do
        tax += delta;
    while (tax > 0 && tax < 100 && getNativeDueLimited(planet, tax, limit) == due);
    sel.setValue(tax);
}

/************************** WColonistTaxSelector *************************/

/** \class WColonistTaxSelector
    Colonist tax selector. This contains the "raw" taxation widget,
    without controls. It displays the income for a particular tax
    rate (default: the planet's colonist tax rate), and allows the
    user to change that if it has focus. */

/** Constructor.
    \param id      widget id
    \param planet  the planet */
WColonistTaxSelector::WColonistTaxSelector(int id, GPlanet& planet, int mifa)
    : UINumberSelector(id, 0, 100, 1), old_value(planet.getColonistTax()), mifa(mifa), planet(planet)
{
    setValue(old_value);
}

void
WColonistTaxSelector::doAutoTax()
{
    setValue(getColonistSafeTax(planet, mifa));
}

bool
WColonistTaxSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (hasState(st_Focused) && e.type == ev_Key) {
        if (e.key.code == ss_Shift + SDLK_LEFT) {
            // Decrease until income changes
            changeColonistTax(*this, planet, -1);
            return true;
        }
        if (e.key.code == ss_Shift + SDLK_RIGHT) {
            // Increase until income changes
            changeColonistTax(*this, planet, +1);
            return true;
        }
        if (e.key.code == '=' || e.key.code == '%') {
            // Set to value
            if (int i = e.key.arg)
                setValue(i);
            return true;
        }
        if (e.key.code == ss_Shift + ' ') {
            // Auto-tax colonists
            doAutoTax();
            return true;
        }
        if (e.key.code == 'u') {
            // Undo
            setValue(planet.getColonistTax());
            return true;
        }
    }
    return UINumberSelector::handleEvent(e, second_pass);
}

void
WColonistTaxSelector::drawContent(GfxCanvas& can)
{
    const int x = extent.x, y = extent.y, w = extent.w,
        h = font_heights[FONT_NORMAL];

    int32 limit;
    const int32 due  = getColonistDueLimited(planet, getValue(), limit);
    const int change = getColonistChange(planet, getValue(), mifa);
    const int new_h  = std::min(100, planet.getColonistHappiness() + getHissEffect(planet, eff)) + change;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();
    ctx.setColor(UIColor::tc_Static);
    outTextF(ctx, x, y,   120, format(_("Tax Rate: %d%%"), getValue()));
    if (planet.getColonistHappiness() < 30 || new_h < host.getPostTaxationHappinessLimit())
        outTextF(ctx, x, y+h, w, format(_("Riots \xE2\x80\x94 Colonists do not pay %d mc."), numToString(due)));
    else
        outTextF(ctx, x, y+h, w, format(_("Colonists pay %d mc."), numToString(due)));
    outTextF(ctx, x, y+2*h, w, format(_("New happiness: %s (%d)"),
                                      getHappinessName(new_h), new_h));
    showChange(ctx, x+120, y, w-120, change);

    can.update(getExtent());
}

void
WColonistTaxSelector::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_FlexibleH;
    info.min_size = info.pref_size = GfxPoint(400, 3 * font_heights[FONT_NORMAL]);
}

/** Commit. Writes back the changed tax rate to the planet. */
void
WColonistTaxSelector::commit()
{
    planet.setColonistTax(mp16_t(getValue()));
}

/** Cancel. This is a hook for future expansion. */
void
WColonistTaxSelector::cancel()
{ }

/*************************** WNativeTaxSelector **************************/

/** \class WNativeTaxSelector
    Native tax selector. This contains the "raw" taxation widget,
    without controls. It displays the income for a particular tax
    rate (default: the planet's native tax rate), and allows the
    user to change that if it has focus.

    The Native tax selector must have an associated colonist tax
    selector, because maximum native tax income depends on colonist
    tax income (if colonists pay N mc, we can collect at most
    MaxPlanetaryIncome-N mc from natives). Note that the caller must
    link these two; we don't make assumptions about the lifetime of
    the colonist tax selector. */

/** Constructor.
    \param id      widget id
    \param planet  the planet
    \param colos   colonist tax selector */
WNativeTaxSelector::WNativeTaxSelector(int id, GPlanet& planet, int mifa, UINumberSelector& colos)
    : UINumberSelector(id, 0, 100, 1), old_value(planet.getNativeTax()), mifa(mifa), planet(planet),
      colonist_tax(colos)
{
    setValue(old_value);
}


void
WNativeTaxSelector::doAutoTax()
{
    setValue(getNativeSafeTax(planet, mifa));
}

void
WNativeTaxSelector::drawContent(GfxCanvas& can)
{
    const int x = extent.x, y = extent.y, w = extent.w,
        h = font_heights[FONT_NORMAL];

    int32 limit;
    const int32 due  = getNativeDue(planet, getValue());
    const int change = getNativeChange(planet, getValue(), mifa);
    const int new_h  = std::min(100, planet.getNativeHappiness() + getHissEffect(planet, eff)) + change;
    const int32 sdue = getBovinoidSupplyContribution(planet);
    const int32 spay = getBovinoidSupplyContributionLimited(planet);

    getColonistDueLimited(planet, colonist_tax.getValue(), limit);
    const int32 pay = getNativeDueLimited(planet, getValue(), limit);

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();
    ctx.setColor(UIColor::tc_Static);
    outTextF(ctx, x, y,   120, format(_("Tax Rate: %d%%"), getValue()));
    outTextF(ctx, x, y+h, w,
             (sdue
              ? format(_("They need to pay %d mc and %d kt supplies."), numToString(due), numToString(sdue))
              : format(_("They need to pay %d mc."), numToString(due))));

    string_t result;
    if (planet.getNativeHappiness() < 30 || new_h < host.getPostTaxationHappinessLimit()) {
        result = _("Riots \xE2\x80\x94 Natives do not pay taxes.");
    } else if (planet.getNativeRace() == nr_Amorphous) {
        result = format(_("Amorphous worms don't pay taxes. They eat %d colonist clan%!1{s%}."),
                        numToString(getAmorphousBreakfast(new_h)));
    } else if (sdue) {
        if (spay < sdue)
            if (pay < due)
                result = format(_("You can collect only %d mc and %d kt supplies."),
                                numToString(pay), numToString(spay));
            else
                result = format(_("You can collect all the money, but only %d kt supplies."),
                                numToString(spay));
        else
            if (pay < due)
                result = format(_("You can collect only %d mc, but all supplies."),
                                numToString(pay));
            else
                result =  _("You can collect all the money and supplies.");
    } else if (pay < due) {
        result = format(_("You can only collect %d mc."), numToString(pay));
    } else {
        result = _("You can collect all the money.");
    }

    drawBackground(ctx, GfxRect(x, y+2*h, w, h*2));
    outTextFormatted(ctx, x, y+2*h, w, result);

    outTextF(ctx, x, y+4*h, w, format(_("New happiness: %s (%d)"),
                                      getHappinessName(new_h), new_h));
    showChange(ctx, x+120, y, w-120, change);

    can.update(getExtent());
}

bool
WNativeTaxSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (hasState(st_Focused) && e.type == ev_Key) {
        if (e.key.code == ss_Shift + SDLK_LEFT) {
            // Decrease until income changes
            changeNativeTax(*this, colonist_tax.getValue(), planet, -1);
            return true;
        }
        if (e.key.code == ss_Shift + SDLK_RIGHT) {
            // Increase until income changes
            changeNativeTax(*this, colonist_tax.getValue(), planet, +1);
            return true;
        }
        if (e.key.code == ss_Shift + ' ') {
            // Auto-tax natives
            doAutoTax();
            return true;
        }
        if (e.key.code == '=' || e.key.code == '%') {
            // Set to value
            if (int i = e.key.arg)
                setValue(i);
            return true;
        }
        if (e.key.code == 'u') {
            // Undo
            setValue(planet.getNativeTax());
            return true;
        }
    }
    return UINumberSelector::handleEvent(e, second_pass);
}

void
WNativeTaxSelector::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_FlexibleH;
    info.min_size = info.pref_size = GfxPoint(400, 5 * font_heights[FONT_NORMAL]);
}

/** Commit. Writes back the changed tax rate to the planet. */
void
WNativeTaxSelector::commit()
{
    planet.setNativeTax(mp16_t(getValue()));
}

/** Cancel. This is a hook for future expansion. */
void
WNativeTaxSelector::cancel()
{ }

/**************************** WTaxationDialog ****************************/

/** \class WTaxationDialog

    Standard taxation dialog. This one displays a WColonistTaxSelector
    and, if needed, a WNativeTaxSelector, all support widgets, wired
    accordingly. */

/** Constructor.
    This builds the complete dialog.
    \param planet   the planet. */
WTaxationDialog::WTaxationDialog(GPlanet& planet, int mi, int fa)
    : UIWindow(_("Taxes"), BLUE_WINDOW, COLOR_GRAY),
      planet(planet),
      ctax(1000, planet, mi+fa),
      p_ntax(0),
      p_growth(0),
      eff(),
      sig_eff()
{
    init(mi, fa);
}

void
WTaxationDialog::init(int mi, int fa)
{
    // Figure out planet effectors
    if (haveDisplayedTurn()) {
        GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
        for (int i = univ.ty_played_ships.findNextIndex(0); i != 0; i = univ.ty_played_ships.findNextIndex(i)) {
            GShip& ship = univ.getShip(i);
            if (ship.isPlayable(GObject::ReadOnly) && ship.getPos() == planet.getPos()) {
                if (raceMissionId(ship.getOwner()) == 2
                    && (ship.getMission() == GMission::msn_Special || ship.getMission() == config.ExtMissionsStartAt() + GMission::pmsn_Special)
                    && ship.getNumBeams() > 0)
                {
                    // Hiss
                    eff.add(GPlanetEffectors::pe_Hiss, 1);
                }
                if (ship.getWaypointDX() == 0 && ship.getWaypointDY() == 0) {
                    // Terraforming is after movement, so only process it if ships have no waypoint.
                    if (ship.canDoSpecial(hf_HeatsTo50)) {
                        eff.add(GPlanetEffectors::pe_Heat50, 1);
                    }
                    if (ship.canDoSpecial(hf_CoolsTo50)) {
                        eff.add(GPlanetEffectors::pe_Cool50, 1);
                    }
                    if (ship.canDoSpecial(hf_HeatsTo100)) {
                        eff.add(GPlanetEffectors::pe_Heat100, 1);
                    }
                }
            }
        }
    }

    setLayout(UIVBoxLayout::instance5);

    // construct the "colonist" part
    {
        UIFocusGroup& fg = h.add(new UIFocusGroup(UIHBoxLayout::instance5, 100, 5));
        UIGroup& v1 = h.add(new UIGroup(UIVBoxLayout::instance5));
        UIGroup& v2 = h.add(new UIGroup(UIVBoxLayout::instance5));

        v1.add(h.add(new UIStaticText(format(_("%s colony:"), player_racenames.getAdjName(planet.getOwner())),
                                      UIStaticText::StaticTitle)));
        v1.add(ctax);
        v2.add(h.add(new UIButton("+", '+')));
        v2.add(h.add(new UIButton("-", '-')));
        v2.add(h.add(new UISpacer()));

        fg.add(v1);
        fg.add(v2);

        add(fg);
        sig_eff.add(&ctax, &WColonistTaxSelector::setEffectors);
    }

    // construct the "natives" part, if needed
    if (planet.getNativeRace()) {
        UIFocusGroup& fg = h.add(new UIFocusGroup(UIHBoxLayout::instance5, 101, 5));
        UIGroup& v1 = h.add(new UIGroup(UIVBoxLayout::instance5));
        UIGroup& v2 = h.add(new UIGroup(UIVBoxLayout::instance5));
        WNativeTaxSelector& ntax = h.add(new WNativeTaxSelector(2000, planet, mi+fa, ctax));
        p_ntax = &ntax;

        v1.add(h.add(new UIStaticText(format(_("%s natives:"), getNativeRaceName(planet.getNativeRace())),
                                      UIStaticText::StaticTitle)));
        v1.add(ntax);
        v2.add(h.add(new UIButton("+", '+')));
        v2.add(h.add(new UIButton("-", '-')));
        v2.add(h.add(new UISpacer()));

        fg.add(v1);
        fg.add(v2);

        add(fg);
        sig_eff.add(&ntax, &WNativeTaxSelector::setEffectors);
        // ntax.setEffectors(eff);

        add(h.add(new UIFocusIterator(100, 101, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Tab)));

        // the "natives" part must be redrawn when the "colonist" part
        // changes, to accomodate for MaxIncome
        ctax.sig_changed.add(this, &WTaxationDialog::redrawNatives);
    }

    // construct the change part
    p_growth = &h.add(new WPlanetGrowthTile(planet));
    add(*p_growth);
    p_growth->setStructures(str_Mines, mi);
    p_growth->setStructures(str_Factories, fa);
    sig_eff.add(p_growth, &WPlanetGrowthTile::setEffectors);
    ctax.sig_changed.add(this, &WTaxationDialog::onChange);
    if (p_ntax)
        p_ntax->sig_changed.add(this, &WTaxationDialog::onChange);

    // publish initial effectors
    sig_eff.raise(eff);

    // buttons
    {
        UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
        UIButton& ok = h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN));
        UIButton& cancel = h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE));
        UIButton& opt = h.add(new UIButton(99, _("Space - Auto Tax"), ' '));
        g.add(ok);
        g.add(cancel);
        g.add(opt);
        g.add(h.add(new UISpacer()));
        g.add(h.add(new UIButton(_("H - Help"), 'h')));
        add(g);
        add(h.add(new WHelpWidget("pcc2:taxes")));

        ok.sig_fired.add(this, &WTaxationDialog::commit);
        cancel.sig_fired.add(this, &WTaxationDialog::cancel);
        opt.sig_fired.add(this, &WTaxationDialog::doAutoTax);
    }

    add(h.add(new UIPrefixArg()));

    pack();
    setState(st_Focused, true);
}

WTaxationDialog::~WTaxationDialog()
{ }

bool
WTaxationDialog::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && e.key.code == vk_Quit) {
        cancel();
        putKey(vk_Quit);
        return true;
    } else if (e.type == ev_Key && e.key.code == 'w') {
        editEffectors();
        return true;
    } else {
        return UIWindow::handleEvent(e, second_pass);
    }
}

/** Redraw native widget. This is used as a slot for sig_changed. */
void
WTaxationDialog::redrawNatives(int)
{
    if (p_ntax)
        p_ntax->drawWidget(false);
}

/** Commit. This is the slot called when the "OK" button is pressed. */
void
WTaxationDialog::commit()
{
    if (p_ntax)
        p_ntax->commit();
    ctax.commit();
    stop(1);
}

/** Cancel. This is the slot called when the "Cancel" button is
    pressed. */
void
WTaxationDialog::cancel()
{
    if (p_ntax)
        p_ntax->cancel();
    ctax.cancel();
    stop(0);
}

void
WTaxationDialog::doAutoTax()
{
    ctax.doAutoTax();
    if (p_ntax)
        p_ntax->doAutoTax();
}

void
WTaxationDialog::onChange()
{
    if (p_growth) {
        p_growth->setColonistTax(ctax.getValue());
        if (p_ntax)
            p_growth->setNativeTax(p_ntax->getValue());
    }
}

void
WTaxationDialog::editEffectors()
{
    GPlanetEffectors oldValues = eff;

    WEffectorDialog dlg(eff, sig_eff);
    dlg.pack();
    ui_root->moveWidgetToEdge(dlg, 2, 2, 10);
    if (ui_root->runModal(dlg) == cm_Escape) {
        eff = oldValues;
        sig_eff.raise(eff);
    }
}

/** Taxation, main entry point. This does the complete taxation dialog for
    planet /p/. */
void
doTaxation(GPlanet& p)
{
    doTaxation(p, p.getStructures(str_Mines), p.getStructures(str_Factories));
}

void
doTaxation(GPlanet& p, int mi, int fa)
{
    WTaxationDialog tax(p, mi, fa);

    ui_root->centerWidget(tax);
    ui_root->add(tax);
    ui_root->run();
    ui_root->close(tax);
}
