/**
  *  \file game/mission.h
  *  \brief Starship Mission Management
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_MISSION_H
#define PCC_V2_GAME_MISSION_H

#include <vector>
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "game/struct.h"
#include "game/playerset.h"

class GShip;
class Stream;

/** Mission description. Describes a ship mission and its arguments.
    A mission is identified by its number and race mask. */
class GMission {
    int        number;           ///< Mission.
    GPlayerSet race_mask;        ///< Race mask.
    int        arg_spec[2];      ///< Argument definitions. ma_XXXX, indexed by Parameter.
    int        flags;            ///< Other flags. mf_XXX.
    string_t   name;             ///< Mission name used in selection lists.
    string_t   short_name;       ///< Short name. Used for scripts.
    string_t   arg_names[2];     ///< Argument names. Indexed by Parameter.
    string_t   exp_condition;    ///< Condition for selectability.
    string_t   exp_warning;      ///< Condition to check whether it will work.
    string_t   exp_label;        ///< Expression for label on ship screen.
    string_t   cmd_onset;        ///< "on set" expression.
    char       hotkey;           ///< Mission hot-key.

    void       parseDescription(const string_t& desc_line);
 public:
    /** Parameter identification for a mission. */
    enum Parameter {
        InterceptParameter,      ///< Intercept parameter.
        TowParameter             ///< Tow parameter.
    };

    GMission();
    GMission(int number, string_t desc_line);
    ~GMission();

    // Mission accessors
    int        getNumber() const;
    GPlayerSet getRaceMask() const;
    void       setRaceMask(GPlayerSet mask);
    int        getFlags() const;
    bool       hasFlag(int i) const;
    void       setFlags(int fl);
    string_t   getName() const;
    void       setName(string_t name);
    string_t   getShortName() const;
    void       setShortName(string_t short_name);
    char       getHotkey() const;
    void       setHotkey(char c);

    // Parameter accessors
    int        getParameterType(Parameter id) const;
    void       setParameterType(Parameter id, int spec);
    string_t   getParameterName(Parameter id) const;
    void       setParameterName(Parameter id, string_t name);

    // Script accessors
    string_t   getConditionExpression() const;
    void       setConditionExpression(string_t cond);
    string_t   getWarningExpression() const;
    void       setWarningExpression(string_t warning);
    string_t   getLabelExpression() const;
    void       setLabelExpression(string_t label);
    string_t   getSetCommand() const;
    void       setSetCommand(string_t cmd);

    // Script execution
    bool       worksOn(GShip& ship) const;
    string_t   getLabel(GShip& ship) const;
    bool       isWarning(GShip& ship) const;

    /** Mission parameter specification. */
    enum {
        ma_TypeMask    = 15,    ///< Type mask.
        ma_Own         = 16,    ///< Own units only.
        ma_NotThis     = 32     ///< This object is not a valid parameter.
    };

    /** Mission argument types (arg_spec & ma_TypeMask).
        Only one of these can be included in a parameter specification. */
    enum {
        mat_Integer    = 1,     ///< Integer.
        mat_PID        = 2,     ///< Planet Id.
        mat_SID        = 3,     ///< Ship Id.
        mat_SID_here   = 4,     ///< Id of a ship here.
        mat_BID        = 5,     ///< Starbase Id.
        mat_Player     = 6      ///< Player number.
    };

    /** Mission flags. */
    enum {
        mf_Waypoint    = 1,     ///< Mission affects waypoint ("Intercept").
        mf_Registered  = 2      ///< Mission is reg-only.
    };

    /** Standard Mission Numbers. */
    enum {
        // Standard missions
        msn_Explore     = 1,
        msn_MineSweep   = 2,
        msn_LayMines    = 3,
        msn_Kill        = 4,
        msn_SensorSweep = 5,
        msn_LnD         = 6,    // Colonize, Land&Disassemble, Decommission
        msn_Tow         = 7,
        msn_Intercept   = 8,
        msn_Special     = 9,
        msn_Cloak       = 10,
        msn_BUFuel      = 11,
        msn_BUDur       = 12,
        msn_BUTri       = 13,
        msn_BUMol       = 14,
        msn_BUSup       = 15,

        // PHost extended missions. These numbers are relative to ExtMissionsStartAt.
        pmsn_BuildTorpsFromCargo = 0,
        pmsn_LayMines            = 1,
        pmsn_LayWeb              = 2,
        pmsn_ScoopTorps          = 3,
        pmsn_GatherBuildTorps    = 4,
        pmsn_BDmc                = 5,
        pmsn_XferTorp            = 6,
        pmsn_XferFtr             = 7,
        pmsn_XferMc              = 8,
        pmsn_StandardSuperSpy    = 9,
        pmsn_Cloak               = 10,
        pmsn_Special             = 11,
        pmsn_GatherBuildFtr      = 12,
        pmsn_BUmc                = 13,
        pmsn_BUclans             = 14,
        pmsn_BeamUpMultiple      = 15,
        pmsn_LayMinesIn          = 16,
        pmsn_LayWebIn            = 17,
        pmsn_Training            = 18
    };
};

/** Mission list. Contains a list of missions. There is a global instance
    as well the possibility to make subset lists. */
class GMissionList {
    typedef std::vector<GMission> container_t;

    container_t data;
    uint32 used_letters;

    static GMissionList instance;
 public:
    typedef container_t::const_iterator iterator;

    GMissionList();
    GMissionList(const GMissionList& other, GShip& ship);
    ~GMissionList();

    /* container interface */
    container_t::size_type getSize() const
        { return data.size(); }
    iterator begin() const
        { return data.begin(); }
    iterator end() const
        { return data.end(); }

    /* manipulator interface */
    bool addMission(const GMission& msn);
    void sort();
    void clear();

    const GMission* getMissionByNumber(int id, GPlayerSet race_mask = GPlayerSet::allUpTo(NUM_PLAYERS)) const;
    const GMission* getMissionByShip(const GShip& ship);

    void loadFromFile(Stream& in);
    void loadFromIniFile(Stream& in);

    /* accessor interface */
    const GMission& operator[](container_t::size_type n) const
        { return data[n]; }

    static GMissionList& getInstance()
        { return instance; }
};

extern const struct ModuleInitializer mod_missions[];

bool isMissionCloaking(int mission_id, int owner);


/******************************** Inlines ********************************/

/** Get mission number. */
inline int
GMission::getNumber() const
{
    return number;
}

/** Get races which can do this mission.
    (This is indeed races, not players!).
    \return set of race numbers */
inline GPlayerSet
GMission::getRaceMask() const
{
    return race_mask;
}

/** Set races which can do this mission. */
inline void
GMission::setRaceMask(GPlayerSet mask)
{
    race_mask = mask;
}

/** Get mission flags. \return mf_XXX */
inline int
GMission::getFlags() const
{
    return flags;
}

/** Check mission flag.
    \param i mf_XXX */
inline bool
GMission::hasFlag(int i) const
{
    return (flags & i) != 0;
}

/** Set mission flags.
    \param fl Set of mf_XXX */
inline void
GMission::setFlags(int fl)
{
    this->flags = fl;
}

/** Get mission name. This name should be displayed in selection lists. */
inline string_t
GMission::getName() const
{
    return name;
}

/** Set mission name. This name is displayed in selection lists. */
inline void
GMission::setName(string_t name)
{
    this->name = name;
}

/** Get short mission name. This name is used when space is tight. */
inline string_t
GMission::getShortName() const
{
    return short_name;
}

/** Set short mission name. This name is used when space is tight. */
inline void
GMission::setShortName(string_t short_name)
{
    this->short_name = short_name;
}

/** Get assigned hot-key.
    The hot-key is an ASCII code (7-bit) for a key to select this mission. */
inline char
GMission::getHotkey() const
{
    return hotkey;
}

/** Assign hot-key.
    The hot-key is an ASCII code (7-bit) for a key to select this mission. */
inline void
GMission::setHotkey(char c)
{
    hotkey = c;
}

/** Get parameter definition.
    \param id which parameter you want to know about
    \return ma_XXX, mat_XXX. Zero if parameter is not required. */
inline int
GMission::getParameterType(Parameter id) const
{
    return arg_spec[id];
}

/** Set parameter type.
    \param id which parameter
    \param spec parameter type, combination of ma_XXX, mat_XXX */
inline void
GMission::setParameterType(Parameter id, int spec)
{
    arg_spec[id] = spec;
}

/** Set parameter name.
    This name is used when PCC2 prompts for a name.
    It can be blank to invoke the default. */
inline void
GMission::setParameterName(Parameter id, string_t name)
{
    arg_names[id] = name;
}

/** Get condition. */
inline string_t
GMission::getConditionExpression() const
{
    return exp_condition;
}

/** Set condition.
    This condition verifies whether the mission is allowed to be set (hard condition).
    \see worksOn(GShip&) */
inline void
GMission::setConditionExpression(string_t cond)
{
    exp_condition = cond;
}

/** Get warning condition. */
inline string_t
GMission::getWarningExpression() const
{
    return exp_warning;
}

/** Set warning condition.
    This condition verifies whether the mission will work (soft condition).
    \see isWarning(GShip&) */
inline void
GMission::setWarningExpression(string_t warning)
{
    exp_warning = warning;
}

/** Get label expression. */
inline string_t
GMission::getLabelExpression() const
{
    return exp_label;
}

/** Set label expression.
    This expression produces a string to display when the mission is set on a ship.
    \see getLabel(GShip&) */
inline void
GMission::setLabelExpression(string_t label)
{
    exp_label = label;
}

/** Get "on set" command. */
inline string_t
GMission::getSetCommand() const
{
    return cmd_onset;
}

/** Set "on set" command. */
inline void
GMission::setSetCommand(string_t cmd)
{
    cmd_onset = cmd;
}

#endif
