/**
  *  \file game/shippredict.cc
  *  \brief Ship Prediction
  *
  *  \todo GShipTurnPredictor operates on TShip objects, not GShip.
  *  This is ugly, as it gives all sorts of problems when requiring
  *  access to ship properties not stored in TShip. However, the
  *  external interface is nice enough, so we don't need to convert
  *  this code until we actually implement the additional
  *  functionality in this module (experience support, per-ship
  *  functions).
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/shippredict.h"
#include "game/univ.h"
#include "game/reg.h"
#include "game/host.h"
#include "game/mission.h"
#include "game/config.h"
#include "game/formula.h"
#include "util/recode.h"
#include "cpluslib/assert.h"

const int GShipTurnPredictor::MOVEMENT_TIME_LIMIT;

enum {
    AlchemyTri = 1,
    AlchemyDur = 2,
    AlchemyMol = 4
};

static int
getAlchemyFCodeValue(const TShip& s, bool enable)
{
    int result = AlchemyTri + AlchemyDur + AlchemyMol;
    if (enable && allowRegistered(s.owner)) {
        if (std::memcmp(s.fcode, "alt", 3) == 0) {
            result = AlchemyTri;
        } else if (std::memcmp(s.fcode, "ald", 3) == 0) {
            result = AlchemyDur;
        } else if (std::memcmp(s.fcode, "alm", 3) == 0) {
            result = AlchemyMol;
        } else if (host.hasAlchemyExclusionFCodes()) {
            if (std::memcmp(s.fcode, "nat", 3) == 0) {
                result -= AlchemyTri;
            } else if (std::memcmp(s.fcode, "nad", 3) == 0) {
                result -= AlchemyDur;
            } else if (std::memcmp(s.fcode, "nam", 3) == 0) {
                result -= AlchemyMol;
            }
        }
    }
    return result;
}

/** Perform refinery reaction.
    \param s   [in/out] Ship
    \param ore [in/out] Ore to consume
    \param sup [in/out] Supplies to consume */
static bool
doRefinery(TShip& s, int16_t& ore, int16_t& sup)
{
    int n = getHull(s.hull).getFuel() - s.ore_cargo[el_Neutronium];
    if (n > ore)
        n = ore;
    if (n > sup)
        n = sup;
    s.ore_cargo[el_Neutronium] += n;
    sup -= n;
    ore -= n;
    return n > 0;
}

/** Perform direct alchemy reaction (Alchemy + Refinery function on one ship).
    \param ship [in/out] Ship
    \param ratio [in] Supply conversion ratio */
static void
doDirectRefinery(TShip& ship, const int ratio, int& used_properties)
{
    int free = getHull(ship.hull).getFuel() - ship.ore_cargo[el_Neutronium];
    int prod = std::min(ship.supplies / ratio, free);
    ship.supplies -= prod*ratio;
    ship.ore_cargo[el_Neutronium] += prod;
    if (prod > 0) {
        used_properties |= GShipTurnPredictor::UsedAlchemy;
    }
}

/** Perform alchemy reaction.
    Merlin converts 3 Supplies -> 1 Mineral.
    Friendly codes allow target mineral control.

    THost and NuHost actually convert 9 Supplies -> 3 Minerals;
    this is what the %rounder is used for. */
static void
doAlchemy(TShip& ship, int& used_properties)
{
    // Merlin converts 3 Sup -> 1 Min
    int t = 0, d = 0, m = 0;
    int mins = ship.supplies / 3;
    int rounder = host.isAlchemyRounding() ? 3 : 1;
    bool used_fc = false;
    switch (getAlchemyFCodeValue(ship, true)) {
     case AlchemyTri:
        t = rounder*(mins/rounder);
        used_fc = true;
        break;

     case AlchemyDur:
        d = rounder*(mins/rounder);
        used_fc = true;
        break;

     case AlchemyMol:
        m = rounder*(mins/rounder);
        used_fc = true;
        break;

     case AlchemyDur + AlchemyMol:
        d = m = mins/2;
        used_fc = true;
        break;

     case AlchemyMol + AlchemyTri:
        t = m = mins/2;
        used_fc = true;
        break;

     case AlchemyTri + AlchemyDur:
        t = d = mins/2;
        used_fc = true;
        break;

     default:
        t = d = m = mins/3;
        break;
    }

    ship.supplies -= 3*(t+d+m);
    ship.ore_cargo[el_Tritanium]  += t;
    ship.ore_cargo[el_Duranium]   += d;
    ship.ore_cargo[el_Molybdenum] += m;

    if (t+d+m > 0) {
        if (used_fc) {
            used_properties |= GShipTurnPredictor::UsedFCode;
        }
        used_properties |= GShipTurnPredictor::UsedAlchemy;
    }
}

static int
getCloakFuel(const TShip& s, int turns, int real_owner)
{
    const int cfb = config.CloakFuelBurn(real_owner);
    int fuel = getHull(s.hull).getMass() * cfb / 100;
    if (fuel < cfb)
        fuel = cfb;
    if (turns)
        fuel *= turns;
    return fuel;
}

static void
normalizePosition(TShip& ship)
{
    GPoint new_pos = GPoint(ship.x, ship.y).getCanonicalLocation();
    ship.x = new_pos.x;
    ship.y = new_pos.y;
}

static int
getEngineLoad(const GUniverse& univ, TShip& ship, TShip* towee_override, bool tow_corr)
{
    int mass = getShipMass(ship);
    if (ship.mission == GMission::msn_Tow && ship.mission_tow > 0 && ship.mission_tow <= NUM_SHIPS) {
        int towee_mass = 0;
        if (towee_override && towee_override->ship_id == ship.mission_tow) {
            towee_mass = getShipMass(*towee_override);
        } else if (univ.ty_any_ships.isValidIndex(ship.mission_tow)) {
            mp16_t t = univ.getShip(ship.mission_tow).getMass();
            if (t.isKnown()) {
                towee_mass = t;
            }
        }
        if (tow_corr)
            towee_mass = 10*(towee_mass/10);
        mass += towee_mass;
    }
    return mass;
}

static int
bitCount8(uint8_t a)
{
    a = (a & 0x55) + ((a & 0xAA) >> 1);
    a = (a & 0x33) + ((a & 0xCC) >> 2);
    return (a & 0x0F) + ((a & 0xF0) >> 4);
}


//   Compute distance, Tim-style.

//   Tim's compiler disagrees with this one (and both differ from IEEE) when
//   considering pythagoras terms. INT(SQR(3^2)) yields 2 with PDS BASIC.

//   Instead of finding a formula which matches that behaviour, we just build
//   a table of distances. Only distances <= 162 ly matter, for other
//   distances it just matters that they're larger. Therefore we don't
//   need to care about precision for large values, and our table has a
//   bounded size.
//
//   See http://phost.de/~stefan/fuelusage.html
static int
timDistance(int dx, int dy)
{
    static const uint8_t TABLE[168][21] = {
        { 253,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255 },
        { 249,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255 },
        { 218,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255 },
        { 233,254,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255 },
        { 177,239,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255 },
        { 210,253,251,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255 },
        {  81,239,255,254,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255 },
        {  97,251,254,255,254,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255 },
        { 161,237,239,255,255,253,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255 },
        { 161,246,254,254,255,255,251,255,255,255,255,255,255,255,255,255,255,255,255,255,255 },
        { 161,186,247,191,255,255,255,223,255,255,255,255,255,255,255,255,255,255,255,255,255 },
        {  66,237,125,255,239,255,255,255,255,254,255,255,255,255,255,255,255,255,255,255,255 },
        {  66,109,239,247,255,251,255,255,255,255,239,255,255,255,255,255,255,255,255,255,255 },
        {  65,181,123,255,254,255,254,255,255,255,255,255,251,255,255,255,255,255,255,255,255 },
        {  65,213,222,247,223,255,255,254,255,255,255,255,255,255,254,255,255,255,255,255,255 },
        {  65, 90,247,126,255,251,255,127,255,255,255,255,255,255,255,255,254,255,255,255,255 },
        {  65,106,187,247,247,127,255,255,255,254,255,255,255,255,255,255,255,255,253,255,255 },
        { 129,170,237,190,127,255,223,255,255,255,253,255,255,255,255,255,255,255,255,255,251 },
        { 129,178,118,239,251,239,255,247,255,255,255,251,255,255,255,255,255,255,255,255,255 },
        { 129,212,218,123,191,255,254,255,251,255,255,255,247,255,255,255,255,255,255,255,255 },
        { 130, 84,235,238,251,251,191,255,255,253,255,255,255,191,255,255,255,255,255,255,255 },
        { 129, 84,109,247,190,191,255,247,255,255,254,255,255,255,255,253,255,255,255,255,255 },
        { 130,164,181,189,247,251,247,255,253,255,127,255,255,255,255,255,239,255,255,255,255 },
        {   2,169,218,238,189,191,255,254,127,255,255,127,255,255,255,255,255,127,255,255,255 },
        {   2,169, 90,119,239,251,251,223,255,223,255,255,127,255,255,255,255,255,255,239,255 },
        {   2,169,106,219,123,223,191,255,251,255,247,255,255,255,254,255,255,255,255,255,255 },
        {   1, 73,173,237,222,251,253,251,127,255,255,251,255,255,255,253,255,255,255,255,255 },
        {   2, 73,213,118,247,222,223,127,255,239,255,255,254,255,255,255,251,255,255,255,255 },
        {   2, 73,213,186,189,247,253,253,239,255,251,255,255,254,255,255,255,247,255,255,255 },
        {   1, 81, 89,219,238,189,223,223,255,253,255,254,255,127,255,255,255,255,223,255,255 },
        {   2, 81,106,109,119,239,253,253,251,191,255,191,255,255,127,255,255,255,255,127,255 },
        {   1,146,170,181,221,123,223,223,191,255,247,255,239,255,255,191,255,255,255,255,255 },
        {   1,146,170,214,238,222,251,254,253,247,255,254,255,247,255,255,127,255,255,255,255 },
        {   1,146,210,218,182,247,222,239,223,255,254,191,255,255,251,255,255,127,255,255,255 },
        {   1,146, 84,107,219,189,247,253,254,253,239,255,239,255,255,253,255,255,255,254,255 },
        {   1,162, 84,173,237,238,190,239,239,191,255,251,255,251,255,255,254,255,255,255,253 },
        {   1, 34, 85,181,118,123,239,253,254,254,251,127,255,255,254,255,127,255,255,255,255 },
        {   1, 34,165,214,186,221,123,223,239,239,127,255,239,255,191,255,255,191,255,255,255 },
        {   1, 34,169, 90,219,238,222,251,254,254,253,239,255,251,255,239,255,255,191,255,255 },
        {   1, 34,169,106,109,187,247,222,239,247,223,255,254,127,255,255,247,255,255,191,255 },
        {   1, 36,169,170,181,221,189,247,125,255,254,253,191,255,223,255,255,251,255,255,191 },
        {   2, 68, 74,181,214,238,238,190,239,247,239,191,255,247,255,247,255,255,253,255,255 },
        {   1, 68, 74,213,218,182,123,239,253,126,255,254,247,255,254,255,253,255,255,254,255 },
        {   1, 68, 82, 85,107,219,221,123,239,247,247,239,255,254,191,255,127,255,255,127,255 },
        {   1, 68, 82,105,173,237,238,222,251,126,127,255,253,223,255,247,255,191,255,255,191 },
        {   2, 68,146,170,181,118,187,247,222,247,247,247,223,255,251,255,253,255,239,255,255 },
        {   1, 68,148,170,214,182,221,189,247,126,127,127,255,251,127,255,127,255,255,247,255 },
        {   2,132,164,170,106,219,110,239,189,239,247,247,239,127,255,239,255,223,255,255,251 },
        {   1,132,164, 82,107,109,183,123,239,125,191,127,255,254,239,255,253,255,247,255,255 },
        {   2,136,164, 84,173,181,219,221,123,239,247,251,247,223,255,253,127,255,255,253,255 },
        {   1,136, 36, 85,181,182,237,246,222,251,190,191,127,255,253,191,255,223,255,127,255 },
        {   2,136, 36, 85,213,218,182,187,247,222,247,251,251,247,191,255,247,255,251,255,223 },
        {   1,136, 40,165, 90,107,219,221,189,247,126,191,191,255,254,247,255,254,255,254,255 },
        {   1,136, 72,169,106,173,237,118,239,189,239,251,251,251,239,255,254,223,255,191,255 },
        {   2,136, 72,169,170,181,110,187,123,239,125,191,191,191,255,253,223,255,247,255,239 },
        {   2,  8, 73,170,172,218,182,221,221,123,239,251,251,251,247,223,255,251,255,254,255 },
        {   1,  8, 73, 74,213, 90,219,110,247,222,251,190,223,191,127,255,251,127,255,191,255 },
        {   2,  8, 73, 82, 85,107,109,183,187,247,222,247,251,253,251,247,127,255,239,255,239 },
        {   2,  8,145, 82, 85,173,181,219,221,189,247,190,223,223,191,255,254,239,255,253,255 },
        {   1,  8,145, 82,105,181,182,109,119,239,189,239,251,253,253,251,239,255,253,127,255 },
        {   2, 16,145,148,170,214,218,182,187,123,239,125,223,223,223,191,255,253,191,255,239 },
        {   2, 16,145,148,170, 90,107,219,221,221,123,239,251,253,253,253,247,191,255,247,255 },
        {   2, 16,146,164,170,106,173,237,118,247,238,123,191,223,223,223,127,255,247,255,254 },
        {   1, 16,146,164, 82,171,181,109,187,187,247,222,247,253,254,253,251,239,255,254,223 },
        {   2, 16, 18,165, 84,181,214,182,221,222,189,247,190,223,239,223,191,255,254,223,255 },
        {   1, 16, 34, 41, 85,213, 90,219,110,119,239,189,247,251,254,254,253,251,223,255,251 },
        {   2, 16, 34, 41, 85, 85,107,109,183,187,123,239,125,223,239,239,223,127,255,251,127 },
        {   1, 16, 34, 73,165,106,173,109,219,237,222,123,239,251,253,254,254,253,247,127,255 },
        {   2, 16, 34, 73,169,170,181,182,237,118,247,238,123,191,239,239,239,223,255,254,239 },
        {   1, 16, 34, 73,170,170,214,218,182,187,187,247,222,247,253,254,254,254,251,239,255 },
        {   1, 16, 34, 82,170,172, 90,107,187,221,238,189,247,190,223,239,239,239,191,255,253 },
        {   1, 32, 36, 82, 74, 85,107,109,221,118,119,239,190,247,253,126,255,254,254,247,191 },
        {   2, 32, 68,146, 82, 85,173,181,109,187,187,123,239,125,223,239,247,239,223,127,255 },
        {   1, 32, 68,146, 82, 85,181,214,182,221,238,222,123,239,251,126,127,255,254,253,239 },
        {   2, 32, 68,146, 84,170,213, 90,219,110,119,247,238,123,223,239,247,247,239,223,255 },
        {   1, 32, 68,162,148,170, 90,107,235,182,187,187,247,222,247,253,126,127,255,254,251 },
        {   1, 32, 68,164,148,170,106,173,109,219,237,238,189,247,190,239,247,247,247,239,191 },
        {   1, 32, 68, 36,165,170,170,181,182,237,118,119,239,189,247,253,126,127,127,255,253 },
        {   1, 32, 68, 36,165, 82,173,214,218,118,187,187,123,239,125,223,247,247,247,247,223 },
        {   1, 32,132, 36,169, 84,213, 90,107,187,237,238,222,123,239,251,126,191,191,127,255 },
        {   1, 32,136, 36, 41, 85, 85,107,109,219,118,119,247,222,123,223,239,247,251,247,239 },
        {   1, 32,136, 68, 73, 85, 85,173,181,109,187,219,187,247,222,251,125,191,191,127,255 },
        {   2, 32,136, 72, 74,165,106,181,214,182,221,238,238,189,247,190,239,247,251,251,247 },
        {   2, 64,136, 72, 82,169,170, 86, 91,219,110,119,119,239,190,247,253,190,191,191,127 },
        {   1, 64,136, 72, 82,170,170, 90,107,219,182,219,189,123,239,189,223,247,251,251,251 },
        {   1, 64,136, 72, 82,170,180,106,173,109,219,237,238,222,123,239,251,126,191,191,191 },
        {   2, 64,  8, 73,146, 74, 85,171,181,182,237,118,119,247,238,123,223,247,251,251,251 },
        {   1, 64,  8,137,148, 82, 85,181,214,218,118,187,221,187,247,222,251,126,191,223,223 },
        {   1, 64,  8,137,148, 84, 85,213, 90,219,182,221,238,238,189,247,190,239,251,251,253 },
        {   1, 64, 16,145,164, 84,169, 86,107,109,219,118,119,119,239,189,247,125,191,223,223 },
        {   1, 64, 16,145,164,148,170, 90,173,181,109,187,219,221,123,239,189,239,247,251,253 },
        {   2, 64, 16,145, 36,165,170,170,213,214,182,219,238,238,222,123,239,251,190,223,223 },
        {   1, 64, 16,145, 36,165,170,170,214,218,218,109,119,119,247,222,123,223,247,251,253 },
        {   1, 64, 16, 17, 41,169, 82,181, 90,107,219,182,219,221,189,247,222,251,126,223,223 },
        {   1, 64, 16, 18, 73, 41, 85, 85,107,173,109,219,237,238,238,189,247,190,239,251,253 },
        {   2, 64, 16, 34, 73, 42, 85, 85,173,182,182,237,118,183,119,239,189,247,125,191,223 },
        {   2, 64, 16, 34, 73, 74, 85, 85,181,214,214,110,187,221,221,123,239,189,239,247,253 },
        {   1,128, 16, 34, 73, 74,165,170,213, 90,219,182,221,238,238,222,123,239,251,190,223 },
        {   2,128, 32, 34, 73, 82,169,170, 90,107,109,219,110,183,123,247,222,123,223,247,251 },
        {   2,128, 32, 34,146, 82,170,170,106,181,181,109,183,219,221,189,247,222,251,190,223 },
        {   1,128, 32, 34,146,146,170,212,170,181,182,182,219,238,238,238,189,247,190,239,251 },
        {   1,128, 32, 34,146,148, 74, 85,173,214,218,182,109,119,187,119,239,189,247,125,223 },
        {   1,128, 32, 36,146,148, 82, 85,213, 90,107,219,182,219,221,221,123,239,189,239,251 },
        {   2,128, 32, 68,146,164, 84, 85, 85,107,173,109,219,237,238,238,222,123,239,123,191 },
        {   1,128, 32, 68,146,164, 84,169, 86,173,181,181,221,118,187,123,247,222,123,223,247 },
        {   1,128, 32, 68, 34, 37,149,170,106,181,214,182,109,187,221,221,189,247,222,251,190 },
        {   1,128, 32, 68, 36, 41,165,170,170,214, 90,219,182,221,238,238,238,189,247,222,247 },
        {   1,128, 32, 68, 36, 41,165,170,170, 90,107,109,219,110,183,187,119,239,189,247,125 },
        {   2,128, 32, 68, 36, 73,169, 82,181,106,173,181,109,183,219,221,221,123,239,189,239 },
        {   2,128, 64, 68, 36, 73, 41, 85, 85,171,181,182,118,219,238,238,238,222,123,239,125 },
        {   1,128, 64,132, 36, 74, 42, 85, 85,181,214,218,182,109,119,187,123,247,222,123,223 },
        {   2,  0, 65,136, 68, 82, 74, 85, 90,213,106,107,219,182,219,221,221,189,247,222,251 },
        {   2,  0, 65,136, 68, 82, 82,165,170, 86,107,173,109,187,237,110,239,238,189,247,222 },
        {   1,  0, 65,136, 72,146, 82,169,170, 90,173,181,181,221,118,187,187,123,239,189,247 },
        {   1,  0, 65,136, 72,146, 82,170,170,170,181,214,182,109,187,221,221,221,123,239,189 },
        {   2,  0, 65,136, 72,146,148,170,212,170, 86, 91,219,182,221,238,118,239,222,123,239 },
        {   1,  0, 65,136, 72,148,148, 74, 85,181, 90,107,109,219,110,183,187,187,247,222,123 },
        {   2,  0, 65,136,136,164,164, 82, 85,213,106,173,173,237,182,219,221,221,189,247,222 },
        {   1,  0, 65,  8,137, 36,165, 84, 85, 85,173,181,182,109,219,237,118,247,238,189,247 },
        {   1,  0, 65,  8,137, 36,165, 84,169, 90,181,214,218,182,237,182,187,187,123,239,189 },
        {   2,  0,129, 16,145, 36, 41,149,170,170,213, 90,107,219,118,219,221,221,221,123,239 },
        {   2,  0,129, 16,145, 36, 41,165,170,170, 90,107,109,109,187,237,110,119,239,222,123 },
        {   1,  0,129, 16,145, 72, 73,169,170,170,106,173,181,109,219,118,187,187,187,247,222 },
        {   2,  0,129, 16, 17, 73, 74,169, 82,213,170,181,214,182,109,187,221,221,221,189,247 },
        {   2,  0,129, 16, 17, 73, 74, 42, 85, 85,173,214, 90,219,182,221,110,119,247,238,189 },
        {   2,  0,129, 16, 18, 73, 82, 74, 85, 85,213, 90,107,107,219,110,183,187,187,123,239 },
        {   1,  0,129, 16, 18, 73, 82, 74, 85, 90, 85,107,173,109,221,182,219,221,221,221,123 },
        {   1,  0,130, 16, 34, 73,146, 82,165,170, 86,173,181,182,109,219,237,118,119,239,222 },
        {   1,  0,130, 16, 34,145,148, 82,169,170,106,181,214,218,182,237,118,187,187,187,247 },
        {   2,  0,130, 32, 34,145,148, 84,170,170,170,214, 90, 91,219,118,219,221,238,221,189 },
        {   1,  0,130, 32, 34,146,164,148,170, 84,171, 90,107,109,107,183,237,110,119,247,238 },
        {   1,  0,  2, 33, 34,146,164,164, 74, 85,181,106,173,181,109,219,118,187,187,187,123 },
        {   2,  0,  2, 33, 34,146, 36,165, 82, 85, 85,171,181,214,182,109,187,221,237,238,221 },
        {   2,  0,  2, 33, 34,146, 36,165, 84, 85, 85,181,214,218,218,182,219,110,119,119,239 },
        {   1,  0,  2, 33, 68, 34, 41, 41, 85,169, 90,213, 90,107,107,219,109,183,187,187,187 },
        {   1,  0,  2, 33, 68, 34, 73, 41,149,170,170, 86,107,173,109,219,182,219,237,238,222 },
        {   1,  0,  2, 33, 68, 34, 73, 73,165,170,170, 90,173,181,182,109,219,237,118,119,247 },
        {   2,  0,  2, 33, 68, 36, 73, 74,169,170,170,170,181,214,218,182,237,118,187,187,187 },
        {   1,  0,  2, 65, 68, 36, 73, 74,169, 82, 85,171,214, 90, 91,219,110,219,221,238,238 },
        {   1,  0,  2, 65, 68, 36, 73, 82, 42, 85, 85,181, 90,107,109,219,182,237,110,119,119 },
        {   1,  0,  2, 65, 68, 36,146, 82, 74, 85, 85,213,106,173,181,109,219,118,187,187,187 },
        {   1,  0,  2, 65,132, 68,146,146, 74, 85,106, 85,173,181,214,182,109,183,221,237,238 },
        {   2,  0,  2, 66,136, 68,146,148, 82,165,170, 90,181,214,218,218,182,219,110,119,119 },
        {   1,  0,  2, 66,136, 68,146,148, 84,170,170,170,213, 90,107,219,186,109,183,187,219 },
        {   2,  0,  4, 66,136, 72,146,164, 84,170,170,170, 86,107,173,109,219,182,219,237,238 },
        {   2,  0,  4, 66,136, 72,162,164,148,170, 84,173,106,173,181,182,109,219,237,118,119 },
        {   1,  0,  4, 66,136, 72, 36, 37,165, 74, 85,213,170,181,214,214,182,221,118,187,219 },
        {   1,  0,  4, 66,136,136, 36, 41,165, 82, 85, 85,173,214, 90,219,218,110,187,221,238 },
        {   2,  0,  4,130,136,136, 36, 41,169, 84, 85, 85,181, 90,107,109,219,182,221,110,119 },
        {   2,  0,  4,130,  8,137, 36, 73, 41, 85,170,106, 85,107,173,181,109,219,110,187,219 },
        {   1,  0,  4,130,  8,137, 36, 73, 41,149,170,170, 86,173,181,182,182,109,183,221,237 },
        {   1,  0,  4,130, 16,145, 68, 74, 74,165,170,170,106,181,214,218,214,118,219,110,119 },
        {   1,  0,  4,130, 16,145, 72, 82, 74,169,170,170,170,213, 90,107,219,182,109,183,187 },
        {   1,  0,  4,130, 16,145, 72, 82, 82,169, 82, 85,171, 90,107,173,109,219,182,219,237 },
        {   2,  0,  4,130, 16, 17, 73,146, 82, 42, 85, 85,181,106,173,181,181,109,187,237,118 },
        {   1,  0,  4,132, 16, 17, 73,146,146, 74, 85, 85, 85,171,181,214,214,182,221,118,187 },
        {   1,  0,  4,132, 16, 17, 73,146,148, 82, 85,170, 85,181,214, 90,219,182,109,187,221 },
        {   1,  0,  4,132, 16, 18,137,148,148, 82,165,170,106,213, 90,107,109,219,182,221,110 },
        {   1,  0,  4,  4, 17, 34,145,164,164, 84,170,170,170, 86,107,173,173,109,219,110,187 },
        {   1,  0,  4,  4, 17, 34,145, 36,165, 84,170,170,170, 90,173,181,182,182,237,182,221 },
        {   1,  0,  4,  4, 33, 34,145, 36, 37,149,170, 84,173,170,181,214,218,182,109,219,110 },
        {   1,  0,  8,  4, 33, 34,146, 36, 41,165, 74, 85, 85,171,214, 90,107,219,182,109,183 },
        {   1,  0,  8,  4, 33, 34,146, 36, 41,165, 82, 85, 85,181, 90,107,109,109,219,118,219 },
        {   1,  0,  8,  4, 33, 34, 18, 73, 73,169, 84, 85, 85,213,106,173,181,181,109,187,237 },
        {   2,  0,  8,  4, 33, 34, 34, 73, 74, 41, 85,170,170, 85,173,181,214,182,182,219,118 },
        {   2,  0,  8,  4, 33, 68, 34, 73, 74, 74,149,170,170, 90,181,214, 90,219,182,109,187 },
        {   2,  0,  8,  4, 33, 68, 34, 73, 82, 74,165,170,170,170,213, 90,107,107,219,182,221 },
        {   2,  0,  8,  8, 33, 68, 36, 73,146, 82,169, 42,181,170, 86,171,173,173,109,219,110 },
    };

    dx = std::abs(dx);
    dy = std::abs(dy);
    if (dy == 0) {
        if (dx == 0) {
            return 0;
        }
        dy = dx;
        dx = 0;
    }
    if (dx >= 168 || dy >= 168) {
        // lo-fi version
        return int(distFromDX(dx, dy));
    }

    int value = dy-1;
    const uint8_t* table_ptr = &TABLE[dy-1][0];
    while (dx >= 8) {
        value += bitCount8(*table_ptr);
        ++table_ptr;
        dx -= 8;
    }

    for (int bit = 0; bit <= dx; ++bit) {
        if ((*table_ptr & (1 << bit)) != 0) {
            ++value;
        }
    }
    return value;
}

static int
computeFuelUsage(const GUniverse& univ, TShip& ship, TShip* towee_override,
                 bool grav_acc, double dist)
{
    if (ship.warp <= 0)
        return 0;

    int way = ship.warp*ship.warp;
    if (grav_acc)
        way *= 2;

    int load = getEngineLoad(univ, ship, towee_override, host.isTHost());
    long ff = getEngine(ship.engine).getFuelFactor(ship.warp);

    if (host.isTHost()) {
        // THost formula
        long move = std::min(timDistance(ship.waypoint_dx, ship.waypoint_dy), way);
        return (ff * (load / 10) * move) / (10000L * way);
    } else if (!config.UseAccurateFuelModel()) {
        // PHost, standard formula
        return rdivaddl(load, 10, 0) * ff * long(dist) / (10000L * way);
    } else {
        // PHost, "accurate" formula
        // This is a cheat. Let's say the formula tells us we need 1.4 kt
        // fuel. If the ship has 5 kt, that'll yield a net usage of 1 kt
        // (5 - 1.4 = 3.6; rounded = 4 kt remaining). If the ship has just
        // one kt, PHost will see that the ship needs more than that, and go
        // into failure mode.
        double res = load * (1 - std::exp(-(ff * dist) / (way*100000L)));
        int res_int = roundToInt(res);
        int ship_fuel = ship.ore_cargo[el_Neutronium];
        if (host.hasAccurateFuelModelBug()) {
            if (res_int == ship_fuel && res > ship_fuel) {
                ++res_int;
            }
        }
        return res_int;
    }
}


/*************************** GShipTurnPredictor **************************/

/** Constructor. Create a predictor for a single ship.
    \param univ universe containing the ship
    \param id   ship Id */
GShipTurnPredictor::GShipTurnPredictor(const GUniverse& univ, int id)
    : id(id), towee_override(0),
      univ(univ), movement_fuel_used(0), cloak_fuel_used(0), turns_computed(0), used_properties(0)
{
    init();
}

/** Constructor. Create a predictor for a ship towing another ship.
    \param univ   universe containing the ship
    \param id     ship Id
    \param towee_override predictor for towed ship */
GShipTurnPredictor::GShipTurnPredictor(const GUniverse& univ, int id, GShipTurnPredictor& towee_override)
    : id(id), towee_override(&towee_override),
      univ(univ), movement_fuel_used(0), cloak_fuel_used(0), turns_computed(0), used_properties(0)
{
    init();
}

void
GShipTurnPredictor::init()
{
    if (univ.getShip(id).hasFullShipData()) {
        univ.getShip(id).getShipData(this->ship);
        ship.owner = univ.getShip(id).getRealOwner();
        valid = true;
    } else {
        valid = false;
    }
}

/** Compute one turn. This will compute all changes for the ship. It
    will also update the towed ship predictor, if any. */
void
GShipTurnPredictor::computeTurn()
{
    // is this actually a predictable ship?
    if (!valid)
        return;

    // where are we?
    int pid                = univ.getPlanetAt(GPoint(ship.x, ship.y));
    const GShip& real_ship = univ.getShip(id);

    // Training
    if (host.isPHost() && ship.mission == config.ExtMissionsStartAt() + GMission::pmsn_Training) {
        ship.warp = 0;
        ship.primary_enemy = 0;
        used_properties |= UsedMission;
    }

    // Mine laying
    // This is a hack: parseMineLayingMission takes a GShip, but we only have a TShip.
    // However, with turns_computed=0, both are equivalent.
    GMineMissionData md;
    if (turns_computed == 0 && parseMineLayingMission(real_ship, univ, md)) {
        ship.ammo -= std::min(int(ship.ammo), md.num_torps);
        if (md.used_fcode) {
            used_properties |= UsedFCode;
        }
        if (md.used_mission) {
            used_properties |= UsedMission;
        }
    }

    // Special Missions I (Super Refit, Self Repair, Hiss, Rob) would go here
    bool is_mkt_fc = std::memcmp(ship.fcode, "mkt", 3) == 0;
    if ((is_mkt_fc
         || (host.isPHost() && ship.mission == config.ExtMissionsStartAt() + GMission::pmsn_BuildTorpsFromCargo))
        && ship.torp_launcher_count > 0 && ship.torp_type > 0 && ship.ore_cargo[el_Neutronium] > 0)
    {
        int torpCost = getTorp(ship.torp_type).getTorpCost();
        int n = ship.ore_cargo[el_Tritanium];
        n = std::min(n, int(ship.ore_cargo[el_Duranium]));
        n = std::min(n, int(ship.ore_cargo[el_Molybdenum]));
        if (torpCost > 0) {
            n = std::min(n, int(ship.money / torpCost));
        }
        ship.ammo += n;
        ship.ore_cargo[el_Tritanium] -= n;
        ship.ore_cargo[el_Duranium] -= n;
        ship.ore_cargo[el_Molybdenum] -= n;
        ship.money -= n * torpCost;

        if (n > 0) {
            if (is_mkt_fc) {
                used_properties |= UsedFCode;
            } else {
                used_properties |= UsedMission;
            }
        }
    }

    /* Alchemy */
    if (std::memcmp(ship.fcode, "NAL", 3) != 0) {
        if (real_ship.canDoSpecial(hf_MerlinAlchemy)) {
            if (host.hasAlchemyCombinations() && config.AllowAdvancedRefinery() && real_ship.canDoSpecial(hf_AriesRefinery)) {
                doDirectRefinery(ship, 3, used_properties);
            } else if (host.hasAlchemyCombinations() && real_ship.canDoSpecial(hf_NeutronicRefinery)) {
                doDirectRefinery(ship, 4, used_properties);
            } else {
                doAlchemy(ship, used_properties);
            }
        } else if (config.AllowAdvancedRefinery() && real_ship.canDoSpecial(hf_AriesRefinery)
                   && (host.hasAlchemyCombinations() || !real_ship.canDoSpecial(hf_NeutronicRefinery)))
        {
            // Aries converts 1 Min -> 1 Neu
            int which = getAlchemyFCodeValue(ship, host.hasRefineryFCodes());
            int16_t max = 0x7FFF;  // we assume that no ship has more cargo
            bool did = false;
            if (which & AlchemyTri) {
                did |= doRefinery(ship, ship.ore_cargo[el_Tritanium], max);
            }
            if (which & AlchemyDur) {
                did |= doRefinery(ship, ship.ore_cargo[el_Duranium], max);
            }
            if (which & AlchemyMol) {
                did |= doRefinery(ship, ship.ore_cargo[el_Molybdenum], max);
            }
            if (did) {
                if (which != AlchemyTri + AlchemyDur + AlchemyMol) {
                    used_properties |= UsedFCode;
                }
                used_properties |= UsedAlchemy;
            }
        } else if (real_ship.canDoSpecial(hf_NeutronicRefinery)) {
            // Neutronic refinery converts 1 Sup + 1 Min -> 1 Neu
            int which = getAlchemyFCodeValue(ship, host.hasRefineryFCodes());
            bool did = false;
            if (which & AlchemyTri) {
                did |= doRefinery(ship, ship.ore_cargo[el_Tritanium], ship.supplies);
            }
            if (which & AlchemyDur) {
                did |= doRefinery(ship, ship.ore_cargo[el_Duranium], ship.supplies);
            }
            if (which & AlchemyMol) {
                did |= doRefinery(ship, ship.ore_cargo[el_Molybdenum], ship.supplies);
            }
            if (did) {
                if (which != AlchemyTri + AlchemyDur + AlchemyMol) {
                    used_properties |= UsedFCode;
                }
                used_properties |= UsedAlchemy;
            }
        }
    } else {
        if (real_ship.canDoSpecial(hf_MerlinAlchemy)
            || real_ship.canDoSpecial(hf_NeutronicRefinery)
            || (config.AllowAdvancedRefinery() && real_ship.canDoSpecial(hf_AriesRefinery)))
        {
            used_properties |= UsedFCode;
        }
    }

    /* Starbase Missions I (Fix, [Recycle, Load Torps]) */
    if (pid > 0) {
        const GPlanet& p = univ.getPlanet(pid);
        if (p.isPlayable(GObject::Playable)
            && p.hasFullBaseData()
            && p.getBaseShipyardOrder() == sya_Fix
            && p.getBaseShipyardId() == id)
        {
            // we're at a base which is fixing us
            ship.damage = 0;
            ship.crew   = getHull(ship.hull).getCrew();
            used_properties |= UsedShipyard;
        }
    }

    /* Supply Repair */
    if (ship.damage > 0 && ship.supplies >= 5) {
        int max = ship.supplies / 5;
        if (max > ship.damage)
            max = ship.damage;
        ship.damage -= max;
        ship.supplies -= 5*max;
        used_properties |= UsedRepair;
    }

    /* Cloak Fuel Burn */
    if ((real_ship.canDoSpecial(hf_Cloak) || real_ship.canDoSpecial(hf_AdvancedCloak))
        && isMissionCloaking(ship.mission, ship.owner))
    {
        int fuel = real_ship.canDoSpecial(hf_AdvancedCloak) ? 0 : getCloakFuel(ship, 0, real_ship.getRealOwner());
        if (ship.ore_cargo[el_Neutronium] <= fuel
            || (ship.damage >= config.DamageLevelForCloakFail()
                && !real_ship.canDoSpecial(hf_HardenedCloak)))
        {
            /* We cancel only cloak missions here. Other missions are
               NOT canceled, see below. */
            ship.mission = 0;
        } else {
            ship.ore_cargo[el_Neutronium] -= fuel;
            cloak_fuel_used += fuel;
            used_properties |= UsedCloak;
        }
    }

    /* This used to check for fuelless ships at this place, and set
       them to mission zero. However, since this is mainly used for
       fuel predictions, this is not a good idea. We want to know how
       much we need, even if you have too little. This mainly affects
       the Tow mission. */

    /* damage speed limit */
    if (ship.damage > 0 && ship.warp > 0 && !real_ship.canDoSpecial(hf_HardenedEngines)) {
        int limit = (raceId(ship.owner) == 2
                     ? (host.isPHost()
                        ? 15 - ship.damage/10
                        : 14 - ship.damage/10)
                     : 10 - ship.damage/10);
        if (limit < 0)
            limit = 0;
        if (ship.warp > limit) {
            ship.warp = limit;
        }

        /* Always set the "damage limit" bit, to confirm that we honored the limit
           even if it didn't end up in an actual restriction */
        used_properties |= UsedDamageLimit;
    }

    /* Actual movement here */
    long dist2 = long(ship.waypoint_dx*ship.waypoint_dx) + long(ship.waypoint_dy*ship.waypoint_dy);
    if (real_ship.canDoSpecial(hf_Hyperdrive)
        && std::memcmp(ship.fcode, "HYP", 3) == 0
        && ship.warp > 0
        && dist2 >= 20*20)   // FIXME: minimum distance not in PHost?
    {
        /* It's hyperjumping */
        ship.ore_cargo[el_Neutronium] -= 50;
        movement_fuel_used            += 50;
        used_properties |= UsedFCode;

        /* If it's jumping, it can't tow. Advance time in towee's
           world anyway. */
        if (ship.mission == GMission::msn_Tow)
            ship.mission = 0;
        if (towee_override)
            towee_override->computeTurn();

        /* Now move that bugger. */
        if (host.isExactHyperjumpDistance2(dist2)) {
            ship.x += ship.waypoint_dx;
            ship.y += ship.waypoint_dy;
        } else {
            // non-exact jump
            double dist = 350 / std::sqrt(double(dist2));
            int mx = int(dist * std::abs(ship.waypoint_dx) + 0.4999999);
            int my = int(dist * std::abs(ship.waypoint_dy) + 0.4999999);
            if (ship.waypoint_dx < 0)
                mx = -mx;
            if (ship.waypoint_dy < 0)
                my = -my;
            ship.x += mx;
            ship.y += my;
        }
        ship.waypoint_dx = ship.waypoint_dy = 0;
        ship.warp = 0;
        normalizePosition(ship);
        // FIXME: gravity wells?
    } else if (dist2 && ship.warp > 0) {
        /* Normal movement */
        // First, compute new position in mx,my
        double dist = std::sqrt(double(dist2));
        int way = ship.warp * ship.warp;
        if (real_ship.canDoSpecial(hf_Gravitonic))
            way *= 2;

        int mx = ship.waypoint_dx, my = ship.waypoint_dy;
        if (dist > way) {
            int dx, dy;
            if (host.isTHost()) {
                // THost movement formulas, from Donovan's
                if (std::abs(mx) > std::abs(my)) {
                    dx = int(double(way) * std::abs(mx) / dist + 0.5);
                    dy = int(double(dx) * std::abs(my) / std::abs(mx) + 0.5);
                } else {
                    dy = int(double(way) * std::abs(my) / dist + 0.5);
                    dx = int(double(dy) * std::abs(mx) / std::abs(my) + 0.5);
                }
                if (mx < 0)
                    dx = -dx;
                if (my < 0)
                    dy = -dy;
            } else {
                // PHost. From docs and source.
                double head = getHeadingRad(mx, my);
                double fx = std::sin(head) * way;
                double fy = std::cos(head) * way;
                dx = int(fx);
                dy = int(fy);
                if (dx != fx)
                    dx += sgn(fx);
                if (dy != fy)
                    dy += sgn(fy);
                if (mx == 0)
                    dx = 0;
                if (my == 0)
                    dy = 0;
            }

            // we now have the dx,dy we want to move
            mx = dx;
            my = dy;
            dist = way;         // FIXME: use distFromDX(dx,dy) instead?
        }

        // Advance time in towee. Must be here because we need to know its "post-movement" mass.
        if (towee_override) {
            if (ship.mission == GMission::msn_Tow) {
                // we assume the tow succeeds. FIXME: be more clever?
                if (towee_override->ship.mission == GMission::msn_Tow)
                    towee_override->ship.mission = 0;
                towee_override->ship.warp = 0;
            }
            towee_override->computeTurn();
            used_properties |= UsedTowee;
        }

        // Compute fuel usage.
        int fuel = computeFuelUsage(univ, ship,
                                    towee_override ? &towee_override->ship : 0,
                                    real_ship.canDoSpecial(hf_Gravitonic),
                                    dist);
        ship.ore_cargo[el_Neutronium] -= fuel;
        movement_fuel_used += fuel;

        // We still have the position offset in mx,my. Move it.
        ship.x += mx;
        ship.y += my;
        ship.waypoint_dx -= mx;
        ship.waypoint_dy -= my;
        normalizePosition(ship);

        // warp wells suck. (the ship to the planet, of course.)
        int wp_x = ship.x + ship.waypoint_dx;
        int wp_y = ship.y + ship.waypoint_dy;

        if (ship.warp > 1 && !univ.getPlanetAt(GPoint(ship.x, ship.y))) {
            if (int gpid = univ.getGravityPlanetAt(GPoint(ship.x, ship.y))) {
                // okay, there is a planet. Move the ship.
                GPoint ppos = univ.getPlanet(gpid).getPos();
                ship.x = ppos.x;
                ship.y = ppos.y;
                // Now adjust the waypoint. If the waypoint was inside
                // the warp well of the same planet, assume the end of
                // this movement order. Otherwise, when users set a
                // waypoint at the edge of the warp well, the ship
                // would try to get there for ever. This is consistent
                // with the CCScript `MoveTo' command and with what
                // most people expect.
                if (univ.getPlanetAt(GPoint(wp_x, wp_y), true) == gpid) {
                    ship.waypoint_dx = ship.waypoint_dy = 0;
                } else {
                    GPoint new_wp = GPoint(wp_x, wp_y).getSimpleNearestAlias(GPoint(ship.x, ship.y));
                    ship.waypoint_dx = new_wp.x - ship.x;
                    ship.waypoint_dy = new_wp.y - ship.y;
                }
            }
        }

        // Update towee position
        if (towee_override) {
            if (ship.mission == GMission::msn_Tow) {
                towee_override->ship.x = ship.x;
                towee_override->ship.y = ship.y;
                towee_override->ship.waypoint_dx = 0;
                towee_override->ship.waypoint_dy = 0;
            }
        }
    } else {
        /* No sensible movement order for this ship. Advance towee's
           time anyway. */
        if (towee_override)
            towee_override->computeTurn();
    }

    // Turn fuel usage.
    int fuel = getTurnFuelUsage(real_ship, config.FuelUsagePerTurnFor100KT, 1);
    if (fuel < ship.ore_cargo[el_Neutronium]) {
        ship.ore_cargo[el_Neutronium] -= fuel;
    } else if (ship.ore_cargo[el_Neutronium] > 0) {
        ship.ore_cargo[el_Neutronium] = 0;
    } else {
        // already dropped below 0. Keep it this way to let them see they consume too much.
    }

    /* turn is over. */
    ++turns_computed;
}

/** Compute this ship's movement, until movement is over or time runs out. */
void
GShipTurnPredictor::computeMovement()
{
    if (valid) {
        const int final_turn = turns_computed + MOVEMENT_TIME_LIMIT;
        while ((ship.waypoint_dx || ship.waypoint_dy) && turns_computed < final_turn) {
            computeTurn();
            if (ship.ore_cargo[el_Neutronium] < 0)
                ship.ore_cargo[el_Neutronium] = 0;
        }
    }
}

/** Override this ship's position. */
void
GShipTurnPredictor::setPosition(GPoint pt)
{
    pt = pt.getSimpleCanonicalLocation();
    ship.x = pt.x;
    ship.y = pt.y;
}

/** Override this ship's waypoint. */
void
GShipTurnPredictor::setWaypoint(GPoint pt)
{
    pt = pt.getSimpleNearestAlias(GPoint(ship.x, ship.y));
    ship.waypoint_dx = pt.x - ship.x;
    ship.waypoint_dy = pt.y - ship.y;
}

/** Override this ship's friendly code. */
void
GShipTurnPredictor::setFCode(string_t s)
{
    storeBasicString(ship.fcode, convertUtf8ToGame(s));
}

/** Get cargo amount. */
int
GShipTurnPredictor::getCargo(GCargoType el) const
{
    switch (el) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        return ship.ore_cargo[el];
     case el_Fighters:
        return ship.bay_count > 0 ? ship.ammo : 0;
     case el_Colonists:
        return ship.colonists;
     case el_Supplies:
        return ship.supplies;
     case el_Money:
        return ship.money;
     default:
        ASSERT(el >= el_Torps && el < el_COUNT);
        if (ship.torp_type == (el - el_Torps+1) && ship.torp_launcher_count > 0) {
            // We know it has this type of torpedoes
            return ship.ammo;
        } else {
            return 0;
        }
    }
}

/** Get this ship's friendly code. */
string_t
GShipTurnPredictor::getFCode() const
{
    return convertGameToUtf8(getBasicString(ship.fcode));
}
