/**
  *  \file game/transact.h
  *  \brief Transactions
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_TRANSACT_H
#define PCC_V2_GAME_TRANSACT_H

#include "game/error.h"
#include "cpluslib/ptrvector.h"

//! Transaction.
class GTransaction {
    GTransaction(const GTransaction&);
    GTransaction& operator=(const GTransaction&);

    // Result tracking
    enum Result { Nothing, Committed, Cancelled };
    Result result      : 8;
    Result must_result : 8;
    void finish(Result);

 protected:
    /** Constructor. */
    GTransaction()
        : result(Nothing), must_result(Nothing)
        { }

    /** Commit the transaction. Derived classes must override this
        method to do the actual work. By convention, a transaction
        should not commit other transactions it did not create
        itself. Instead, the transaction user should use a
        GMultiTransaction to set them up as a group and commit them
        as a group. For consistency checking, you can call
        mustCommit() on the related transactions to make sure they
        are committed, not cancelled.
        \pre !getError() */
    virtual void commit() = 0;

    /** Cancel (rollback) the transaction. Derived classes must
        override this method to do the actual rollback. The convention
        described for commit() applies here, too. You can use
        mustCancel() to mark a transaction as needing to be
        cancelled. */
    virtual void cancel() = 0;

 public:
    virtual ~GTransaction();

    void doCommit();
    void doCancel();
    bool isFinished() const;
    bool isCommitted() const;
    bool isCancelled() const;
    void mustCommit();
    void mustCancel();

    virtual GError getError();
};

//! Multi-Part Transaction.
class GMultiTransaction : public GTransaction {
 protected:
    void commit();
    void cancel();

 public:
    GMultiTransaction();
    ~GMultiTransaction();

    template<class T> T& addNewSubtransaction(T* t);

    virtual GError getError();

 private:
    ptr_vector<GTransaction> children;
};


/** Check whether this transaction was finished cleanly (commit or cancel). */
inline bool
GTransaction::isFinished() const
{
    return result != Nothing;
}

/** Check whether this transaction was committed (i.e.\ it is neither unfinished nor cancelled). */
inline bool
GTransaction::isCommitted() const
{
    return result == Committed;
}

/** Check whether this transaction was cancelled (i.e.\ it is neither unfinished nor committed). */
inline bool
GTransaction::isCancelled() const
{
    return result == Cancelled;
}

/** Add newly-allocated subtransaction. Use as
    <tt>MyTA& m = mpt.addNewSubtransaction(new MyTA())</tt>. */
template<class T>
inline T&
GMultiTransaction::addNewSubtransaction(T* t)
{
    children.push_back(t);
    return *t;
}

#endif
