/**
  *  \file gfx/gfxcompl.cc
  *  \brief Complex Graphics Primitives
  *
  *  Graphics primitives built atop the others. Hence clipping
  *  algorithms need not deal with clipping a circle or a general
  *  line.
  *
  *  In addition, this includes graphics primitives that are just
  *  wrappers with a more convenient interface.
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <cmath>
#include "gfx/gfx.h"
#include "gfx/gfxcompl.h"
#include "gfx/gfxcontext.h"

#define LINE_PATTERN(parm, t) ((parm).line_style & (0x80 >> (t & 7)))
#define LINE_PATTERN1(parm, t) ((parm).getLinePattern() & (0x80 >> (t & 7)))

/// Helper function for 1-wide lines.
static void
line_pset(GfxContext& c, int x1, int y1)
{
    c.getCanvas().drawPixel(GfxPoint(x1, y1), c.getRawColor(), c.getAlpha());
}

/// Helper function for general lines in north-south direction.
static void
line_hline(GfxContext& c, int x1, int y1)
{
    c.getCanvas().drawHLine(GfxPoint(x1, y1), c.getLineThickness(), c.getRawColor(), SOLID_LINE, c.getAlpha());
}

/// Helper function for general lines in east-west direction.
static void
line_vline(GfxContext& c, int x1, int y1)
{
    c.getCanvas().drawVLine(GfxPoint(x1, y1), c.getLineThickness(), c.getRawColor(), SOLID_LINE, c.getAlpha());
}

/** \defgroup GfxCompl Complex Graphics Primitives

    These graphics primitives are built atop the other primitives.
    They need not, and can not, be overridden in GfxCanvas' descendants.

    They take their parameters from a GfxContext object. */
//@{

/** Draw Horizontal Line. Draw a horizontal line from (x1,y1) to (x2,y1),
    inclusive. There are no restrictions regarding the relations between
    the ending points.
    \param ctx parameters
    \param x1,y1 starting coordinates
    \param x2 ending X. */
void
drawHLine(GfxContext& ctx, int x1, int y1, int x2)
{
    if (x1 > x2)
        std::swap(x1, x2);
    x2++;

    if (ctx.getLineThickness() == 1) {
        ctx.getCanvas().drawHLine(GfxPoint(x1, y1), x2-x1, ctx.getRawColor(), ctx.getLinePattern(), ctx.getAlpha());
    } else {
        GfxFillPattern pat(ctx.getLinePattern());
        GfxRect r(x1,    y1 - ctx.getLineThickness()/2,
                  x2-x1, ctx.getLineThickness());
        ctx.getCanvas().drawBar(r, ctx.getRawColor(), GFX_TRANSPARENT, pat, ctx.getAlpha());
    }
}

/** Draw Vertical Line. Draw a vertical line from (x1,y1) to (x1,y2),
    inclusive. There are no restrictions regarding the relations between
    the ending points.
    \param ctx parameters
    \param x1,y1 starting coordinates
    \param y2 ending Y */
void
drawVLine(GfxContext& ctx, int x1, int y1, int y2)
{
    if (y2 < y1)
        std::swap(y1, y2);
    y2++;

    if (ctx.getLineThickness() == 1) {
        ctx.getCanvas().drawVLine(GfxPoint(x1, y1), y2-y1, ctx.getRawColor(), ctx.getLinePattern(), ctx.getAlpha());
    } else {
        GfxFillPattern pat;
        for (int i = 0; i < 8; ++i)
            if ((ctx.getLinePattern() << i) & 0x80)
                pat[i] = 0xFF;
        GfxRect r(x1 - ctx.getLineThickness()/2, y1,
                  ctx.getLineThickness(),        y2-y1);
        ctx.getCanvas().drawBar(r, ctx.getRawColor(), GFX_TRANSPARENT, pat, ctx.getAlpha());
    }
}

/** Draw Line. Draws a general line from (x1,y1) to (x2,y2), inclusive.
    This is a standard Bresenham algorithm, with shortcuts for
    horizontal and vertical lines. There are no restrictions
    regarding the relations between the endpoints.
    \param ctx parameters
    \param p1 starting coordinates
    \param p2 ending coordinates */
void
drawLine(GfxContext& ctx, const GfxPoint p1, const GfxPoint p2)
{
    int x1 = p1.x, y1 = p1.y;
    int x2 = p2.x, y2 = p2.y;

    /* Handle easy cases first */
    if (x1 == x2) {
        drawVLine(ctx, x1, y1, y2);
        return;
    }
    if (y1 == y2) {
        drawHLine(ctx, x1, y1, x2);
        return;
    }

    /* Vanilla Bresenham */
    int dx = x2 - x1;
    int dy = y2 - y1;
    if(dx < 0)
        dx = -dx;
    if(dy < 0)
        dy = -dy;

    if(dx < dy) {                // Y is the major axis
        if(y1 > y2) {
            std::swap(y1, y2);
            std::swap(x1, x2);
        }

        void (*plot) (GfxContext&,int,int);
        int error = dy >> 1;
        int addx  = (x2 < x1) ? -1 : 1;

        if(ctx.getLineThickness() == 1) {
            plot = line_pset;
        } else {
            /* FIXME: correct error term for even line thicknesses */
            x1 -= ctx.getLineThickness() / 2;
            plot = line_hline;
        }

        while(y1 <= y2) {
            if (LINE_PATTERN1(ctx, y1))
                plot(ctx, x1, y1);
            y1++;
            error += dx;
            if(error >= dy) {
                error -= dy;
                x1 += addx;
            }
        }
    } else {                // X is the major axis
        if(x1 > x2) {
            std::swap(y1, y2);
            std::swap(x1, x2);
        }

        void (*plot) (GfxContext&,int,int);
        int error = dx >> 1;
        int addy  = (y2<y1) ? -1 : 1;

        if(ctx.getLineThickness() == 1) {
            plot = line_pset;
        } else {
            /* FIXME: correct error term for even line thicknesses */
            y1 -= ctx.getLineThickness() / 2;
            plot = line_vline;
        }

        while(x1 <= x2) {
            if (LINE_PATTERN1(ctx, x1))
                plot(ctx, x1, y1);
            x1++;
            error += dy;
            if(error >= dx) {
                error -= dx;
                y1 += addy;
            }
        }
    }
}

/** Draw Line to. Draws a line between the current graphics cursor in
    \c parm to the specified (x,y) coordinates. Otherwise like drawLine(),
    i.e. endpoints are inclusive.
    \param ctx parameters. All line styles and the graphics cursor are used,
               the cursor is updated.
    \param x,y endpoint*/
void
drawLineTo(GfxContext& ctx, const GfxPoint pt)
{
    drawLine(ctx, ctx.getCursor(), pt);
    ctx.setCursor(pt);
}


/** Draw Relative Line. Draws a line from the current graphics cursor in
    \c parm to the point offset by (dx,dy) from it. Otherwise like
    drawLine().
    \param ctx parameters. All line styles and the graphics cursor are used,
               the cursor is updated.
    \param dx,dy offset of endpoint relative to cursor. */
void
drawLineRel(GfxContext& ctx, int dx, int dy)
{
    drawLineTo(ctx, ctx.getCursor().movedBy(dx, dy));
}


/** Circle. Draws a circle around (x0,y0), with the specified radius.

    The algorithm variation used here draws pixels that are less than
    r+1 from center.

    \todo some pixels in the diagonals are painted twice which is a
    Bad Thing(tm) for alpha circles

    \param ctx parameters. All line styles are used.
    \param pt center point
    \param r radius */
void
drawCircle(GfxContext& ctx, const GfxPoint pt, int r)
{
    /*
     *  A home-brew circle algorithm (but probably not at all new):
     *  . it uses the equation sum(1,3,5,...,2n-1) = n^2 to compute
     *    pythagoras' theorem.
     *  . I first implemented it in Z80 assembly language
     *  If you have a better algorithm, tell me.
     */
    const int x0 = pt.x;
    const int y0 = pt.y;
#if 1
    /* FIXME: This is the circle routine used in PCC I, which does not
       support line-patterned circles and is quite slow. The other one
       below has different (=wrong) rounding characteristics. We want
       a circle of radius 3 to appear as a "rounded rectangle" with
       diameter 7, the other draws squares with diameter 5 instead. */
    register int w = 0;
    register int k = 1;
    register int z = 0;
    register int y2 = 0;
    register int xk = 2*r-1;
    register int x = r;
    GfxCanvas& can = ctx.getCanvas();
    uint32_t color = ctx.getRawColor();
    alpha_type alpha = ctx.getAlpha();
    while (x >= 0) {
        while (z < y2) {
            z += k;
            ++w;
            k += 2;
        }
        if (w >= x) {
            can.drawPixel(GfxPoint(x0 - x, y0 - w), color, alpha);
            can.drawPixel(GfxPoint(x0 + x, y0 - w), color, alpha);
            can.drawPixel(GfxPoint(x0 - w, y0 - x), color, alpha);
            can.drawPixel(GfxPoint(x0 + w, y0 - x), color, alpha);
            can.drawPixel(GfxPoint(x0 - w, y0 + x), color, alpha);
            can.drawPixel(GfxPoint(x0 + w, y0 + x), color, alpha);
            can.drawPixel(GfxPoint(x0 - x, y0 + w), color, alpha);
            can.drawPixel(GfxPoint(x0 + x, y0 + w), color, alpha);
        }
        y2 += xk;
        xk -= 2;
        --x;
    }
#else
    register int w = 0;
    register int k = 1;
    register int z = 0;
    register int y2 = 0;
    register int xk = 2*r-1;
    register int x = r;

    register int w0 = w;
    while(x >= w) {
        while(z < y2) {
            z += k;
            drawVLine(ctx, x0+x, y0+w0, y0+w);
            drawVLine(ctx, x0-x, y0+w0, y0+w);
            drawHLine(ctx, x0+w0, y0+x, x0+w);
            drawHLine(ctx, x0+w0, y0-x, x0+w);
            if (w) {
                drawVLine(ctx, x0+x, y0-w0, y0-w);
                drawVLine(ctx, x0-x, y0-w0, y0-w);
                drawHLine(ctx, x0-w0, y0+x, x0-w);
                drawHLine(ctx, x0-w0, y0-x, x0-w);
            }
            w++;
            w0 = w;
            k += 2;
        }
        y2 += xk;
        xk -= 2;
        x--;
    }
#endif
}

/** Draw filled circle. The circle is filled with the current fill settings.

    \param ctx parameters
    \param pt center point
    \param r radius */
void
drawFilledCircle(GfxContext& ctx, GfxPoint pt, int r)
{
    /* Roughly the same algorithm as drawCircle */
    const int x0 = pt.x;
    const int y0 = pt.y;
    register int w = 0;
    register int k = 1;
    register int z = 0;
    register int y2 = 0;
    register int xk = 2*r-1;
    register int x = r;

    register int w0 = w;
    while(x >= 0) {
        while(z < y2) {
            z += k;
            drawBar(ctx, x0 - x, y0 + w0, x0 + x, y0 + w);
            if (w)
                drawBar(ctx, x0 - x, y0 - w0, x0 + x, y0 - w);
            w++;
            w0 = w;
            k += 2;
        }
        y2 += xk;
        xk -= 2;
        x--;
    }
}

/** Draw Filled Bar. Draws a filled bar between the corners (x1,y1)
    and (x2,y2), inclusive. The given points can be any two opposing
    corners of the rectangle.
    \param ctx parameters. All fill styles are used.
    \param x1,y1 one corner
    \param x2,y2 other corner */
void
drawBar(GfxContext& ctx, int x1, int y1, int x2, int y2)
{
    if (x1 > x2)
        std::swap(x1, x2);
    if (y1 > y2)
        std::swap(y1, y2);

    GfxRect r(x1, y1, x2-x1+1, y2-y1+1);
    ctx.getCanvas().drawBar(r, ctx.getRawColor(), GFX_TRANSPARENT, ctx.getFillPattern(), ctx.getAlpha());
}

/** Draw Filled Bar. Fills rectangle r with color and fill pattern
    from \c parm. Note that when you know alpha and color, you can
    also call drawBar() directly.
    \param ctx parameters. All fill styles are used.
    \param r rectangle to fill */
void
drawBar(GfxContext& ctx, const GfxRect& r)
{
    ctx.getCanvas().drawBar(r, ctx.getRawColor(), GFX_TRANSPARENT, ctx.getFillPattern(), ctx.getAlpha());
}

/** Draw Solid Bar. Draws a solid rectangle with the given color. This
    one does not use a graphics parameter set.
    \param can canvas
    \param r rectangle to fill
    \param color color to fill the rectangle with */
void
drawSolidBar(GfxCanvas& can, const GfxRect& r, uint32 color)
{
    can.drawBar(r, color, GFX_TRANSPARENT, SOLID_FILL, GFX_ALPHA_OPAQUE);
}

void
drawSolidBar(GfxContext& ctx, const GfxRect& r, uint32_t color)
{
    ctx.getCanvas().drawBar(r, ctx.getColorScheme().getColor(color), GFX_TRANSPARENT, SOLID_FILL, ctx.getAlpha());
}

/** Draw rectangle. Draw the borders of a rectangle, but does not fill
    the object. Degenerate rectangles (size 0 or 1) are supported.
    \param ctx parameters. All line styles are used.
    \param r rectangle to draw */
void
drawRectangle(GfxContext& ctx, const GfxRect& r)
{
    if (r.w == 0 || r.h == 0) {
        return;
    } else if (r.w == 1) {
        drawVLine(ctx, r.x, r.y, r.y + r.h - 1);
    } else if (r.h == 1) {
        drawHLine(ctx, r.x, r.y, r.x + r.w - 1);
    } else {
        int x2 = r.x + r.w - 1;
        int y2 = r.y + r.h - 1;
        drawHLine(ctx, r.x, r.y, x2);
        drawHLine(ctx, r.x,  y2, x2);
        drawVLine(ctx, r.x, r.y+1, y2-1);
        drawVLine(ctx, x2,  r.y+1, y2-1);
    }
}

static inline int32 square(int32 x) { return x*x; }

/** Draw an arrow.
    \param ctx       drawing parameters
    \param p1        starting coordinates
    \param p2        ending coordinates (the pointer is added here)
    \param ptsize    length of pointer in pixels, approximately */
void
drawArrow(GfxContext& ctx, const GfxPoint p1, const GfxPoint p2, int ptsize)
{
    drawLine(ctx, p1, p2);

    int l = int(std::sqrt(double(square(p2.x - p1.x) + square(p2.y - p1.y))));
    if (l) {
        int dx = (p2.x - p1.x) * ptsize / l;
        int dy = (p2.y - p1.y) * ptsize / l;
        drawLine(ctx, p2, GfxPoint(p2.x-dy-dx, p2.y+dx-dy));
        drawLine(ctx, p2, GfxPoint(p2.x+dy-dx, p2.y-dx-dy));
    }
}

/** Draw pixel.
    \param ctx graphics context
    \param pt point */
void
drawPixel(GfxContext& ctx, const GfxPoint pt)
{
    ctx.getCanvas().drawPixel(pt, ctx.getRawColor(), ctx.getAlpha());
}

/** Draw background.
    \param ctx graphics context
    \param pt area */
void
drawBackground(GfxContext& ctx, const GfxRect& r)
{
    ctx.getColorScheme().drawBackground(ctx, r);
}

//@}
