#!/usr/bin/perl -w
#
#  I/O Table Generator
#
#  Usage: iogen BASENAME
#
#  Reads BASENAME.in, a file containing structure definitions.
#
#  Writes BASENAME.h containing the same structure definitions, augmented
#  with 'size', 'count' and 'parser' members for marshalling.
#
#  Writes BASENAME.cc containing the marshalling tables.
#
#  Input syntax (not free-form, linefeeds do matter!):
#     struct NAME {               Structure definition
#        TYPE NAME[DIM],....;     Member definition
#     };
#     struct NAME;                Declares that NAME is a known structure
#
#  TYPE can be
#     int16, int16_t
#     uint16, uint16_t
#     int32, int32_t
#     uint32, uint32_t
#     char, uint8, uint8_t, int8
#     Name of a structure defined or declared previously
#  Lines can contain the "attribute" RECODED to request a string to be recoded
#  from external to internal charset upon load, as in
#     struct foo {
#        char name[20] RECODED;
#     };
#
use strict;

my $basename = shift @ARGV or die "usage: $0 <basename>";
my $intype = "";
my $offset = 0;
my $offsetadd = "";
my $length = 0;
my %sizeof;
my %sizeadd;
my %external;
open IN, "<$basename.in" or die "can't open $basename.in: $!";
open H, ">$basename.h" or die "can't create $basename.h: $!";
open CC, ">$basename.cc" or die "can't create $basename.cc: $!";
my $had_blurb = 0;
my $ingroup = "";
my $exit = 0;
while(<IN>) {
    chomp;
    if($intype ne "") {
	if(/\};/) {
	    print CC "    { T_END, 0, 0, 0, 0 }\n";
	    print CC "};\n";
            print CC "CTASSERT(countof(", $intype, "::parser) == ", $length+1, ");\n\n";
            my $op = $offsetadd ne '' ? ' + ' : '';
	    print H "    enum {\n";
	    print H "       size  = $offset$op$offsetadd,      ///< Size of this structure in DOS\n";
	    print H "       count = $length       ///< Number of members\n";
            print H "    };\n";
	    print H "    static const TStructureParser parser\[\];\n";
	    $sizeof{$intype} = $offset;
            $sizeadd{$intype} = $offsetadd;
	    $intype = "";
        } elsif(m{^\s*//} || m{^\s*$}) {
            # empty
        } elsif(/^(\s*\/\*.*)/) {
            while (! /\*\//) {
                print H "$_\n";
                chomp ($_ = <IN>);
            }
	} elsif(/^(\s*)((?:const\s*|mutable\s*)*)(\w+)(\s+)([^;]+);(.*)/) {
            my ($ws1, $mod, $type, $ws2, $rest, $com) = ($1, $2, $3, $4, $5, $6);

            # process all "pragmas" here
            my $recoded = ($rest =~ s/\bRECODED\b//);
            my $ignored = ($rest =~ s/\bIGNORED\b//); # for future extension

            # update line
            $rest =~ s/\s+$//;
            $_ = "$ws1$mod$type$ws2$rest;$com";

	    my @values = split /, */, $rest;
            print STDERR "warning: RECODED specified but item is not type char\n"
                if $recoded && $type ne "char";
	    foreach (@values) {
		my ($name, $number);
		$name = $_;
		$number = 1;
		while ($name =~ s/\s*\[(\d+)\]$//) {
		    $number *= $1;
		}
		if ($name !~ /^\w+$/) {
		    print STDERR "warning: name `$name' looks wrong\n";
		}
		if($type eq "int16" || $type eq "int16_t") {
		    print CC "    { T_INT16, $number, offsetof($intype, $name), 0, 0 },\n";
		    $offset += $number * 2;
		    $length++;
                } elsif($type eq "uint16" || $type eq "uint16_t") {
		    print CC "    { T_UINT16, $number, offsetof($intype, $name), 0, 0 },\n";
		    $offset += $number * 2;
		    $length++;
		} elsif($type eq "int32" || $type eq "int32_t") {
		    print CC "    { T_INT32, $number, offsetof($intype, $name), 0, 0 },\n";
		    $offset += $number * 4;
		    $length++;
		} elsif($type eq "uint32" || $type eq "uint32_t") {
		    print CC "    { T_UINT32, $number, offsetof($intype, $name), 0, 0 },\n";
		    $offset += $number * 4;
		    $length++;
		} elsif($type eq "char" || $type eq "uint8" || $type eq "uint8_t" || $type eq 'int8') {
                    print CC "    { T_STRING, $number, offsetof($intype, $name), 0, 0 },\n";
		    $offset += $number;
		    $length++;
		} elsif(defined $sizeof{$type}) {
		    print CC "    { T_TYPEREF, $number, offsetof($intype, $name), sizeof($type), ${type}::parser },\n";
		    $offset += $sizeof{$type} * $number;
                    if ($sizeadd{$type} ne '') {
                        $offsetadd .= " + " if $offsetadd ne '';
                        if ($number == 1) {
                            $offsetadd .= "$sizeadd{$type}";
                        } else {
                            if ($sizeadd{$type} =~ / /) {
                                $offsetadd .= "($sizeadd{$type}) * $number";
                            } else {
                                $offsetadd .= "$sizeadd{$type} * $number";
                            }
                        }
                    }
		    $length++;
                } elsif(defined $external{$type}) {
                    print CC "    { T_TYPEREF, $number, offsetof($intype, $name), sizeof($type), ${type}::parser },\n";
                    $offsetadd .= " + " if $offsetadd ne '';
                    if ($number == 1) {
                        $offsetadd .= $type."::size";
                    } else {
                        $offsetadd .= $type."::size * $number";
                    }
		    $length++;
		} else {
		    print STDERR "can't parse: `$type $_'\n";
                    $exit = 1;
		}
	    }
	} else {
	    print STDERR "can't parse: `$_'\n";
            $exit = 1;
	}
    } else {
        if (/\\ingroup\s+(.*)/) { 
            $ingroup = "  *  \\ingroup $1\n"; 
        } elsif (!$had_blurb) {
            print CC "/**
  *  \\file $basename.cc
  *  \\brief Structure Definitions (Conversion Information)
$ingroup  *
  *  Automatically generated by $0 from $basename.in
  */\n\n";
            print CC "#include \"io/io.h\"\n";
            print CC "#include \"cpluslib/misc.h\"\n";
            print CC "#include \"cpluslib/assert.h\"\n";
            print CC "#include \"$basename.h\"\n\n";
            print H "/**
  *  \\file $basename.h
  *  \\brief Structure Definitions (Structures)
$ingroup  *
  *  Automatically generated by $0 from $basename.in
  */\n\n";
            $had_blurb = 1;
        }
	if(/struct (\w+) \{/) {
	    $intype = $1;
	    $offset = 0;
            $offsetadd = '';
	    $length = 0;
	    print CC "/** Conversion information. Describes how to convert a\n";
	    print CC "    $intype structure to and from DOS. */\n";
	    print CC "const TStructureParser ${intype}::parser\[] = {\n";
	}
        if (/struct (\w+);/) {
            $external{$1} = 1;
        }
    }
    print H $_, "\n";
}
exit $exit;
