/**
  *  \file u/t_int_builtin.cc
  *  \brief Tests for builtin functions
  *
  *  Untested:
  *      CC$Trace -- This is a debug feature that has been tried during debugging :-)
  *      Count -- Needs an indexable context mock
  *      Find -- Needs an indexable context mock
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "u/t_int.h"

/** Test trig functions: Sin, Cos, Tan, ATan.
    Instructions: ucos, usin, utan, batan */
void
IntBuiltinTestSuite::testTrig()
{
    // Sin
    checkFloatExpr("sin(0)", 0);
    checkFloatExpr("sin(90)", 1);
    checkFloatExpr("sin(180)", 0);
    checkFloatExpr("sin(270)", -1);
    checkFloatExpr("sin(0.0)", 0);
    checkFloatExpr("sin(90.0)", 1);
    checkFloatExpr("sin(180.0)", 0);
    checkFloatExpr("sin(270.0)", -1);
    checkNullExpr("sin(z(0))");
    checkExecFailure("sin('a')");
    checkExecFailure("sin('')");

    // Cos
    checkFloatExpr("cos(0)", 1);
    checkFloatExpr("cos(90)", 0);
    checkFloatExpr("cos(180)", -1);
    checkFloatExpr("cos(270)", 0);
    checkFloatExpr("cos(0.0)", 1);
    checkFloatExpr("cos(90.0)", 0);
    checkFloatExpr("cos(180.0)", -1);
    checkFloatExpr("cos(270.0)", 0);
    checkNullExpr("cos(z(0))");
    checkExecFailure("cos('a')");
    checkExecFailure("cos('')");

    // Tan
    checkFloatExpr("tan(0)", 0);
    checkFloatExpr("tan(45)", 1);
    checkExecFailure("tan(90)");
    checkFloatExpr("tan(135)", -1);
    checkFloatExpr("tan(180)", 0);
    checkExecFailure("tan(270)");
    checkFloatExpr("tan(0.0)", 0);
    checkFloatExpr("tan(45.0)", 1);
    checkExecFailure("tan(90.0)");
    checkFloatExpr("tan(135.0)", -1);
    checkFloatExpr("tan(180.0)", 0);
    checkExecFailure("tan(270.0)");
    checkNullExpr("tan(z(0))");
    checkExecFailure("tan('a')");
    checkExecFailure("tan('')");

    // Atan
    checkFloatExpr("atan(1)", 45);
    checkFloatExpr("atan(-1)", 315);
    checkFloatExpr("atan(1.0)", 45);
    checkFloatExpr("atan(-1.0)", 315);

    checkFloatExpr("atan(1,1)", 45);
    checkFloatExpr("atan(-1,1)", 315);
    checkFloatExpr("atan(1,-1)", 135);
    checkFloatExpr("atan(-1,-1)", 225);

    checkFloatExpr("atan(1.0,1.0)", 45);
    checkFloatExpr("atan(-1.0,1.0)", 315);
    checkFloatExpr("atan(1.0,-1.0)", 135);
    checkFloatExpr("atan(-1.0,-1.0)", 225);

    checkFloatExpr("atan(0)", 0);
    checkFloatExpr("atan(1, 0)", 90);
    checkFloatExpr("atan(0, 1)", 0);
    checkFloatExpr("atan(-1, 0.0)", 270);
    checkFloatExpr("atan(0, -1.0)", 180);

    checkFloatExpr("atan(sin(25), cos(25))", 25);

    checkNullExpr("atan(1, z(0))");
    checkNullExpr("atan(z(0))");
    checkNullExpr("atan(z(0), 1)");

    checkExecFailure("atan('a')");
    checkExecFailure("atan(1,'a')");
    checkExecFailure("atan('a',1)");
}

/** Test Abs function.
    Instruction: uabs */
void
IntBuiltinTestSuite::testAbs()
{
    checkIntExpr("abs(1)", 1);
    checkIntExpr("abs(-1)", 1);
    checkIntExpr("abs(true)", 1);
    checkIntExpr("abs(0)", 0);
    checkIntExpr("abs(-1111111111)", 1111111111);
    checkIntExpr("abs(+1111111111)", 1111111111);

    checkFloatExpr("abs(1.0)", 1.0);
    checkFloatExpr("abs(-1.0)", 1.0);
    checkFloatExpr("abs(-9999.0)", 9999.0);
    checkFloatExpr("abs(pi)", 3.14159265);
    checkFloatExpr("abs(-pi)", 3.14159265);
    checkFloatExpr("abs(7777777.0)", 7777777.0);

    checkNullExpr("abs(z(0))");

    checkExecFailure("abs('a')");
    checkExecFailure("abs('')");
}

/** Test Asc function.
    Instruction: uasc */
void
IntBuiltinTestSuite::testAsc()
{
    checkIntExpr("asc('a')", 97);
    checkIntExpr("asc('abcdef')", 97);
    checkIntExpr("asc(0)", 48);
    checkIntExpr("asc(2.5)", 50);
    checkNullExpr("asc('')");
    checkNullExpr("asc(z(0))");
}

/** Test bit operations.
    Instructions: bbitand, bbitor, bbitxor, ubitnot */
void
IntBuiltinTestSuite::testBitOps()
{
    // BitAnd
    checkIntExpr("bitand(1)", 1);
    checkIntExpr("bitand(7,85)", 5);
    checkIntExpr("bitand(4,4,4)", 4);
    checkIntExpr("bitand(4,4,0)", 0);
    checkIntExpr("bitand(true,true)", 1);
    checkNullExpr("bitand(z(0))");
    checkNullExpr("bitand(z(0),1)");
    checkNullExpr("bitand(1,z(0))");
    checkNullExpr("bitand(1,1,z(0))");
    checkNullExpr("bitand(z(0),1,1)");
    checkExecFailure("bitand('a')");
    checkExecFailure("bitand(1,'a')");
    checkExecFailure("bitand('a',1)");

    // BitOr
    checkIntExpr("bitor(1)", 1);
    checkIntExpr("bitor(7,85)", 87);
    checkIntExpr("bitor(4,4,4)", 4);
    checkIntExpr("bitor(4,4,0)", 4);
    checkIntExpr("bitor(true,false)", 1);
    checkNullExpr("bitor(z(0))");
    checkNullExpr("bitor(z(0),1)");
    checkNullExpr("bitor(1,z(0))");
    checkNullExpr("bitor(1,1,z(0))");
    checkNullExpr("bitor(z(0),1,1)");
    checkExecFailure("bitor('a')");
    checkExecFailure("bitor(1,'a')");
    checkExecFailure("bitor('a',1)");

    // BitXor
    checkIntExpr("bitxor(1)", 1);
    checkIntExpr("bitxor(7,85)", 82);
    checkIntExpr("bitxor(4,4,4)", 4);
    checkIntExpr("bitxor(4,4,0)", 0);
    checkIntExpr("bitxor(true,true)", 0);
    checkNullExpr("bitxor(z(0))");
    checkNullExpr("bitxor(z(0),1)");
    checkNullExpr("bitxor(1,z(0))");
    checkNullExpr("bitxor(1,1,z(0))");
    checkNullExpr("bitxor(z(0),1,1)");
    checkExecFailure("bitxor('a')");
    checkExecFailure("bitxor(1,'a')");
    checkExecFailure("bitxor('a',1)");

    // BitNot
    checkIntExpr("bitnot(1)", -2);
    checkIntExpr("bitnot(0)", -1);
    checkIntExpr("bitnot(true)", -2);
    checkNullExpr("bitnot(z(0))");
    checkExecFailure("bitnot('a')");
}

/** Test Min/Max operators.
    Instructions: bmin, bmax, and NC versions thereof */
void
IntBuiltinTestSuite::testMinMax()
{
    // Integers
    checkIntExpr("min(1)", 1);
    checkIntExpr("min(3,1,4,1,5)", 1);
    checkIntExpr("min(99,-22,22)", -22);
    checkIntExpr("max(1)", 1);
    checkIntExpr("max(3,1,4,1,5)", 5);
    checkIntExpr("max(99,-22,22)", 99);
    checkNullExpr("min(1,z(0))");
    checkNullExpr("min(z(0),1)");
    checkNullExpr("min(z(0))");
    checkNullExpr("max(1,z(0))");
    checkNullExpr("max(z(0),1)");
    checkNullExpr("max(z(0))");

    // Floats and Mixes
    // Note that min/max do not lift their argument to a common type before returning it!
    checkFloatExpr("min(1.0)", 1);
    checkFloatExpr("min(3.1, 4.1, 5.9, 2.6)", 2.6);
    checkIntExpr("min(99.0,-22,22)", -22);
    checkFloatExpr("max(1.0)", 1);
    checkFloatExpr("max(3.1, 4.1, 5.9, 2.6)", 5.9);
    checkFloatExpr("max(99.0,-22,22)", 99);
    checkNullExpr("min(1.0,z(0))");
    checkNullExpr("min(z(0),1.0)");
    checkNullExpr("min(z(0))");
    checkNullExpr("max(1.0,z(0))");
    checkNullExpr("max(z(0),1.0)");
    checkNullExpr("max(z(0))");

    // Same things with StrCase, to exercise NC versions
    checkIntExpr("strcase(min(1))", 1);
    checkIntExpr("strcase(min(3,1,4,1,5))", 1);
    checkIntExpr("strcase(min(99,-22,22))", -22);
    checkIntExpr("strcase(max(1))", 1);
    checkIntExpr("strcase(max(3,1,4,1,5))", 5);
    checkIntExpr("strcase(max(99,-22,22))", 99);

    checkFloatExpr("strcase(min(1.0))", 1);
    checkFloatExpr("strcase(min(3.1, 4.1, 5.9, 2.6))", 2.6);
    checkIntExpr("strcase(min(99.0,-22,22))", -22);
    checkFloatExpr("strcase(max(1.0))", 1);
    checkFloatExpr("strcase(max(3.1, 4.1, 5.9, 2.6))", 5.9);
    checkFloatExpr("strcase(max(99.0,-22,22))", 99);

    checkNullExpr("strcase(min(1.0,z(0)))");
    checkNullExpr("strcase(min(z(0),1.0))");
    checkNullExpr("strcase(min(z(0)))");
    checkNullExpr("strcase(max(1.0,z(0)))");
    checkNullExpr("strcase(max(z(0),1.0))");
    checkNullExpr("strcase(max(z(0)))");
    
    // Strings
    checkStringExpr("min('h','a','l','l','o')", "a");
    checkStringExpr("max('h','a','l','l','o')", "o");
    checkStringExpr("min('H','a','L','l','O')", "a");
    checkStringExpr("max('H','a','L','l','O')", "O");

    checkStringExpr("strcase(min('h','a','l','l','o'))", "a");
    checkStringExpr("strcase(max('h','a','l','l','o'))", "o");
    checkStringExpr("strcase(min('H','a','L','l','O'))", "H");
    checkStringExpr("strcase(max('H','a','L','l','O'))", "l");

    checkNullExpr("min('a',z(0))");
    checkNullExpr("min(z(0),'a')");
    checkNullExpr("strcase(min('a',z(0)))");
    checkNullExpr("strcase(min(z(0),'a'))");
}

/** Test Chr/Chr$ function (two names for the same function).
    Instructions: uchr */
void
IntBuiltinTestSuite::testChr()
{
    checkStringExpr("chr(1)", "\001");
    checkStringExpr("chr(97)", "a");
    checkNullExpr("chr(z(0))");
    checkIntExpr("len(chr(0))", 1);
    checkExecFailure("chr('a')");
    checkStringExpr("chr(128)", "\xC2\x80");
    checkStringExpr("chr(57665)", "\xEE\x85\x81");

    checkStringExpr("chr$(1)", "\001");
    checkStringExpr("chr$(97)", "a");
    checkNullExpr("chr$(z(0))");
    checkIntExpr("len(chr$(0))", 1);
    checkExecFailure("chr$('a')");
    checkStringExpr("chr$(128)", "\xC2\x80");
    checkStringExpr("chr$(57665)", "\xEE\x85\x81");
}

/** Test type tests: IsEmpty, IsNum, IsString.
    Instructions: uisempty, uisnum, uisstr */
void
IntBuiltinTestSuite::testTypeChecks()
{
    checkBoolExpr("isempty(0)", false);
    checkBoolExpr("isempty(1)", false);
    checkBoolExpr("isempty(0.0)", false);
    checkBoolExpr("isempty(1.0)", false);
    checkBoolExpr("isempty(true)", false);
    checkBoolExpr("isempty(false)", false);
    checkBoolExpr("isempty('')", false);
    checkBoolExpr("isempty('foo')", false);
    checkBoolExpr("isempty(z(0))", true);
    checkBoolExpr("isempty(zap(0))", true);

    checkBoolExpr("isnum(0)", true);
    checkBoolExpr("isnum(1)", true);
    checkBoolExpr("isnum(0.0)", true);
    checkBoolExpr("isnum(1.0)", true);
    checkBoolExpr("isnum(true)", true);
    checkBoolExpr("isnum(false)", true);
    checkBoolExpr("isnum('')", false);
    checkBoolExpr("isnum('foo')", false);
    checkBoolExpr("isnum(z(0))", false);
    checkBoolExpr("isnum(zap(0))", false);

    checkBoolExpr("isstring(0)", false);
    checkBoolExpr("isstring(1)", false);
    checkBoolExpr("isstring(0.0)", false);
    checkBoolExpr("isstring(1.0)", false);
    checkBoolExpr("isstring(true)", false);
    checkBoolExpr("isstring(false)", false);
    checkBoolExpr("isstring('')", true);
    checkBoolExpr("isstring('foo')", true);
    checkBoolExpr("isstring(z(0))", false);
    checkBoolExpr("isstring(zap(0))", false);
}

/** Test Exp/Log.
    Instructions: uexp, ulog */
void
IntBuiltinTestSuite::testExp()
{
    checkFloatExpr("exp(-1)", 1/2.7172);
    checkFloatExpr("exp(0)", 1);
    checkFloatExpr("exp(1)", 2.7172);
    checkFloatExpr("exp(2)", 2.7172*2.7172);

    checkFloatExpr("exp(-1.0)", 1/2.7172);
    checkFloatExpr("exp(0.0)", 1);
    checkFloatExpr("exp(1.0)", 2.7172);
    checkFloatExpr("exp(2.0)", 2.7172*2.7172);

    checkFloatExpr("log(2.7172*2.7172)", 2);
    checkFloatExpr("log(2.7172)", 1);
    checkFloatExpr("log(1)", 0);

    checkFloatExpr("exp(log(10000))", 10000);
    checkFloatExpr("log(exp(10))", 10);
    // This fails because exp(10000) overflows, yielding NaN. CxxTest then fails
    // to convert the offending value into a string. D'oh.
    // checkFloatExpr("log(exp(10000))", 10000);

    checkExecFailure("log(0)");
    checkExecFailure("log('a')");
    checkExecFailure("exp('a')");
    checkNullExpr("log(z(0))");
    checkNullExpr("exp(z(0))");
}

/** Test substring-finding routines: InStr, First, Rest.
    Instructions: bfindstr, bfirststr, breststr, and NC versions thereof. */
void
IntBuiltinTestSuite::testStrFind()
{
    // InStr
    checkIntExpr("instr('foobar', 'o')", 2);
    checkIntExpr("instr('foobar', 'O')", 2);
    checkIntExpr("instr('foobar', 'oO')", 2);
    checkIntExpr("instr('quuxUUM', 'UU')", 2);
    checkIntExpr("instr('foo', 'bar')", 0);
    checkIntExpr("instr('foobar', 'foo')", 1);
    checkIntExpr("instr('foo', 'foobar')", 0);
    checkIntExpr("instr('foobar', 'b')", 4);

    checkIntExpr("strcase(instr('foobar', 'o'))", 2);
    checkIntExpr("strcase(instr('foobar', 'O'))", 0);
    checkIntExpr("strcase(instr('foobar', 'oO'))", 0);
    checkIntExpr("strcase(instr('quuxUUM', 'UU'))", 5);
    checkIntExpr("strcase(instr('foo', 'bar'))", 0);
    checkIntExpr("strcase(instr('foobar', 'foo'))", 1);
    checkIntExpr("strcase(instr('foo', 'foobar'))", 0);
    checkIntExpr("strcase(instr('foobar', 'b'))", 4);

    // First - different order of parameters!
    checkStringExpr("first('o', 'foobar')", "f");
    checkStringExpr("first('O', 'foobar')", "f");
    checkStringExpr("first('oO', 'foobar')", "f");
    checkStringExpr("first('UU', 'quuxUUM')", "q");
    checkStringExpr("first('bar', 'foo')", "foo");
    checkStringExpr("first('foo', 'foobar')", "");
    checkStringExpr("first('foobar', 'foo')", "foo");
    checkStringExpr("first('b', 'foobar')", "foo");

    checkStringExpr("strcase(first('o', 'foobar'))", "f");
    checkStringExpr("strcase(first('O', 'foobar'))", "foobar");
    checkStringExpr("strcase(first('oO', 'foobar'))", "foobar");
    checkStringExpr("strcase(first('UU', 'quuxUUM'))", "quux");
    checkStringExpr("strcase(first('bar', 'foo'))", "foo");
    checkStringExpr("strcase(first('foo', 'foobar'))", "");
    checkStringExpr("strcase(first('foobar', 'foo'))", "foo");
    checkStringExpr("strcase(first('b', 'foobar'))", "foo");

    // Rest
    checkStringExpr("rest('o', 'foobar')", "obar");
    checkStringExpr("rest('O', 'foobar')", "obar");
    checkStringExpr("rest('oO', 'foobar')", "bar");
    checkStringExpr("rest('UU', 'quuxUUM')", "xUUM");
    checkNullExpr("rest('bar', 'foo')");
    checkStringExpr("rest('foo', 'foobar')", "bar");
    checkNullExpr("rest('foobar', 'foo')");
    checkStringExpr("rest('b', 'foobar')", "ar");

    checkStringExpr("strcase(rest('o', 'foobar'))", "obar");
    checkNullExpr("strcase(rest('O', 'foobar'))");
    checkNullExpr("strcase(rest('oO', 'foobar'))");
    checkStringExpr("strcase(rest('UU', 'quuxUUM'))", "M");
    checkNullExpr("strcase(rest('bar', 'foo'))");
    checkStringExpr("strcase(rest('foo', 'foobar'))", "bar");
    checkNullExpr("strcase(rest('foobar', 'foo'))");
    checkStringExpr("strcase(rest('b', 'foobar'))", "ar");

    // Null
    checkNullExpr("instr(z(0),'a')");
    checkNullExpr("instr('a',z(0))");
    checkNullExpr("instr(z(0),z(0))");
    checkNullExpr("first(z(0),'a')");
    checkNullExpr("first('a',z(0))");
    checkNullExpr("first(z(0),z(0))");
    checkNullExpr("rest(z(0),'a')");
    checkNullExpr("rest('a',z(0))");
    checkNullExpr("rest(z(0),z(0))");

    // Type errors
    // FIXME: should these really be type errors, or should we implicitly stringify?
    checkExecFailure("instr('a', 1)");
    checkExecFailure("instr(1, 'a')");
    checkExecFailure("instr(1, 2)");
    checkExecFailure("first('a', 1)");
    checkExecFailure("first(1, 'a')");
    checkExecFailure("first(1, 2)");
    checkExecFailure("rest('a', 1)");
    checkExecFailure("rest(1, 'a')");
    checkExecFailure("rest(1, 2)");

}

/** Test substring operations: Mid, Left, Right.
    Instructions: blcut, brcut, bendcut */
void
IntBuiltinTestSuite::testSubstr()
{
    // Mid, 2-arg
    checkStringExpr("mid('foobar',-2)", "foobar");
    checkStringExpr("mid('foobar',-1)", "foobar");
    checkStringExpr("mid('foobar',0)", "foobar");
    checkStringExpr("mid('foobar',1)", "foobar");
    checkStringExpr("mid('foobar',2)", "oobar");
    checkStringExpr("mid('foobar',3)", "obar");
    checkStringExpr("mid('foobar',4)", "bar");
    checkStringExpr("mid('foobar',5)", "ar");
    checkStringExpr("mid('foobar',6)", "r");
    checkStringExpr("mid('foobar',7)", "");
    checkStringExpr("mid('foobar',8)", "");

    checkNullExpr("mid(z(0),1)");
    checkNullExpr("mid(z(0),z(0))");
    checkNullExpr("mid('foo',z(0))");

    // Mid, 3-arg
    checkStringExpr("mid('foobar',-2,3)", "foo");
    checkStringExpr("mid('foobar',-1,3)", "foo");
    checkStringExpr("mid('foobar',0,3)", "foo");
    checkStringExpr("mid('foobar',1,3)", "foo");
    checkStringExpr("mid('foobar',2,3)", "oob");
    checkStringExpr("mid('foobar',3,3)", "oba");
    checkStringExpr("mid('foobar',4,3)", "bar");
    checkStringExpr("mid('foobar',5,3)", "ar");
    checkStringExpr("mid('foobar',6,3)", "r");
    checkStringExpr("mid('foobar',7,3)", "");
    checkStringExpr("mid('foobar',8,3)", "");

    checkStringExpr("mid('foobar',2,0)", "");

    checkNullExpr("mid(z(0),1,1)");
    checkNullExpr("mid(z(0),z(0),1)");
    checkNullExpr("mid('foo',z(0),1)");
    checkNullExpr("mid(z(0),1,z(0))");
    checkNullExpr("mid(z(0),z(0),z(0))");
    checkNullExpr("mid('foo',z(0),z(0))");
    checkNullExpr("mid('foo',1,z(0))");
    
    // Left
    checkStringExpr("left('foobar',-3)", "");
    checkStringExpr("left('foobar',-2)", "");
    checkStringExpr("left('foobar',-1)", "");
    checkStringExpr("left('foobar',0)", "");
    checkStringExpr("left('foobar',1)", "f");
    checkStringExpr("left('foobar',2)", "fo");
    checkStringExpr("left('foobar',3)", "foo");
    checkStringExpr("left('foobar',4)", "foob");
    checkStringExpr("left('foobar',5)", "fooba");
    checkStringExpr("left('foobar',6)", "foobar");
    checkStringExpr("left('foobar',7)", "foobar");
    checkStringExpr("left('foobar',8)", "foobar");

    checkNullExpr("left(z(0),1)");
    checkNullExpr("left(z(0),z(0))");
    checkNullExpr("left('foo',z(0))");

    // Right
    checkStringExpr("right('foobar',-3)", "");
    checkStringExpr("right('foobar',-2)", "");
    checkStringExpr("right('foobar',-1)", "");
    checkStringExpr("right('foobar',0)", "");
    checkStringExpr("right('foobar',1)", "r");
    checkStringExpr("right('foobar',2)", "ar");
    checkStringExpr("right('foobar',3)", "bar");
    checkStringExpr("right('foobar',4)", "obar");
    checkStringExpr("right('foobar',5)", "oobar");
    checkStringExpr("right('foobar',6)", "foobar");
    checkStringExpr("right('foobar',7)", "foobar");
    checkStringExpr("right('foobar',8)", "foobar");

    checkNullExpr("right(z(0),1)");
    checkNullExpr("right(z(0),z(0))");
    checkNullExpr("right('foo',z(0))");

    // Type errors
    checkExecFailure("mid(10,1,1)");
    checkExecFailure("mid('foo','1','1')");
    checkExecFailure("mid('foo','1')");

    checkExecFailure("left(10, 1)");
    checkExecFailure("left('foo', '1')");

    checkExecFailure("right(10, 1)");
    checkExecFailure("right('foo', '1')");
}

/** Test trim functions: Trim, LTrim, RTrim.
    Instructions: ulrtrim, ultrim, urtrim */
void
IntBuiltinTestSuite::testTrim()
{
    // Trim
    checkStringExpr("trim('foo')", "foo");
    checkStringExpr("trim('  foo')", "foo");
    checkStringExpr("trim('foo  ')", "foo");
    checkStringExpr("trim('  foo  ')", "foo");
    checkStringExpr("trim('    ')", "");
    checkNullExpr("trim(z(0))");
    checkExecFailure("trim(1)");
    checkExecFailure("trim(1.0)");

    // LTrim
    checkStringExpr("ltrim('foo')", "foo");
    checkStringExpr("ltrim('  foo')", "foo");
    checkStringExpr("ltrim('foo  ')", "foo  ");
    checkStringExpr("ltrim('  foo  ')", "foo  ");
    checkStringExpr("ltrim('    ')", "");
    checkNullExpr("ltrim(z(0))");
    checkExecFailure("ltrim(1)");
    checkExecFailure("ltrim(1.0)");

    // RTrim
    checkStringExpr("rtrim('foo')", "foo");
    checkStringExpr("rtrim('  foo')", "  foo");
    checkStringExpr("rtrim('foo  ')", "foo");
    checkStringExpr("rtrim('  foo  ')", "  foo");
    checkStringExpr("rtrim('    ')", "");
    checkNullExpr("rtrim(z(0))");
    checkExecFailure("rtrim(1)");
    checkExecFailure("rtrim(1.0)");
}

/** Test square root (Sqr/Sqrt).
    Instructions: usqrt */
void
IntBuiltinTestSuite::testSqrt()
{
    // Sqrt
    checkFloatExpr("sqrt(0)", 0);
    checkFloatExpr("sqrt(1)", 1);
    checkFloatExpr("sqrt(2)", 1.4142);
    checkFloatExpr("sqrt(3)", 1.73205);
    checkFloatExpr("sqrt(4)", 2);

    checkFloatExpr("sqrt(0.0)", 0);
    checkFloatExpr("sqrt(1.0)", 1);
    checkFloatExpr("sqrt(2.0)", 1.4142);
    checkFloatExpr("sqrt(3.0)", 1.73205);
    checkFloatExpr("sqrt(4.0)", 2);

    checkFloatExpr("sqrt(1.23456^2)", 1.23456);
    checkFloatExpr("sqrt(12345)^2", 12345);

    checkNullExpr("sqrt(z(0))");
    checkNullExpr("sqrt(z(0)^2)");

    checkExecFailure("sqrt(-1)");
    checkExecFailure("sqrt('a')");

    // Same things again with Sqr
    checkFloatExpr("sqr(0)", 0);
    checkFloatExpr("sqr(1)", 1);
    checkFloatExpr("sqr(2)", 1.4142);
    checkFloatExpr("sqr(3)", 1.73205);
    checkFloatExpr("sqr(4)", 2);

    checkFloatExpr("sqr(0.0)", 0);
    checkFloatExpr("sqr(1.0)", 1);
    checkFloatExpr("sqr(2.0)", 1.4142);
    checkFloatExpr("sqr(3.0)", 1.73205);
    checkFloatExpr("sqr(4.0)", 2);

    checkFloatExpr("sqr(1.23456^2)", 1.23456);
    checkFloatExpr("sqr(12345)^2", 12345);

    checkNullExpr("sqr(z(0))");
    checkNullExpr("sqr(z(0)^2)");

    checkExecFailure("sqr(-1)");
    checkExecFailure("sqr('a')");
}

/** Test rounding functions: Int, Round.
    Instructions: utrunc, uround */
void
IntBuiltinTestSuite::testRound()
{
    // Integers
    checkIntExpr("int(-1)", -1);
    checkIntExpr("int(0)", 0);
    checkIntExpr("int(+1)", 1);
    checkIntExpr("int(true)", 1);
    checkIntExpr("round(-1)", -1);
    checkIntExpr("round(0)", 0);
    checkIntExpr("round(+1)", 1);
    checkIntExpr("round(true)", 1);

    // Floats, Int
    checkIntExpr("int(1.9)", 1);
    checkIntExpr("int(1.5)", 1);
    checkIntExpr("int(1.1)", 1);
    checkIntExpr("int(0.4)", 0);
    checkIntExpr("int(0.0)", 0);
    checkIntExpr("int(-0.4)", 0);
    checkIntExpr("int(-1.1)", -1);
    checkIntExpr("int(-1.5)", -1);
    checkIntExpr("int(-1.9)", -1);
    checkIntExpr("int(2147483647)", 2147483647);
    checkNullExpr("int(z(0))");
    checkExecFailure("int('a')");
    checkExecFailure("int(2147483648)");

    // Floats, Round
    checkIntExpr("round(1.9)", 2);
    checkIntExpr("round(1.5)", 2);
    checkIntExpr("round(1.1)", 1);
    checkIntExpr("round(0.4)", 0);
    checkIntExpr("round(0.0)", 0);
    checkIntExpr("round(-0.4)", 0);
    checkIntExpr("round(-1.1)", -1);
    checkIntExpr("round(-1.5)", -2);
    checkIntExpr("round(-1.9)", -2);
    checkIntExpr("round(2147483647)", 2147483647);
    checkNullExpr("round(z(0))");
    checkExecFailure("round('a')");
    checkExecFailure("round(2147483648)");
}

/** Test If function.
    Several other tests can be found in IntParseExprTestSuite::testAnd etc. */
void
IntBuiltinTestSuite::testIf()
{
    checkIntExpr("if(1,2,3)", 2);
    checkIntExpr("if(0,2,3)", 3);
    checkNullExpr("if(0,2)");
    checkIntExpr("if(1,2)", 2);

    checkIntExpr("if(1,2,'a')",2);
    checkStringExpr("if(0,2,'a')", "a");
}

/** Test Str function.
    Instructions: ustr, bstr */
void
IntBuiltinTestSuite::testStr()
{
    // Unary
    checkStringExpr("str(1)", "1");
    checkStringExpr("str(123456789)", "123456789");
    checkStringExpr("str(1.0)", "1");
    checkStringExpr("str(1.01)", "1.01");
    // FIXME: the following holds for PCC1, but not for PCC2:
    // checkStringExpr("str(1.001)", "1");
    checkStringExpr("str('a')", "a");
    checkStringExpr("str(true)", "YES");
    checkStringExpr("str(false)", "NO");
    checkNullExpr("str(z(0))");

    // Binary
    checkStringExpr("str(1,3)", "1.000");
    checkStringExpr("str(123456789,3)", "123456789.000");
    checkStringExpr("str(1.0,3)", "1.000");
    checkStringExpr("str(1.01,3)", "1.010");
    checkStringExpr("str(1.001,3)", "1.001");
    checkExecFailure("str('a',3)");
    checkStringExpr("str(true,3)", "YES");
    checkStringExpr("str(false,3)", "NO");
    checkNullExpr("str(z(0),3)");
    checkNullExpr("str(1,z(0))");
    checkExecFailure("str(1, 'a')");
}

/** Test Val function.
    Instruction: uval */
void
IntBuiltinTestSuite::testVal()
{
    checkIntExpr("val('1')", 1);
    checkIntExpr("val('99')", 99);
    checkIntExpr("val('-99')", -99);
    checkIntExpr("val('   1 ')", 1);
    checkFloatExpr("val('1.0')", 1);
    checkFloatExpr("val('99.0')", 99);
    checkFloatExpr("val('-99.0')", -99);
    checkFloatExpr("val('.5')", 0.5);
    checkFloatExpr("val('1.')", 1.0);
    checkFloatExpr("val('   1.0 ')", 1);

    checkNullExpr("val('')");
    checkNullExpr("val('a')");

    checkExecFailure("val(1)");
    checkExecFailure("val(true)");
}

/** Test Z/Zap function.
    Instruction: uzap */
void
IntBuiltinTestSuite::testZap()
{
    // Null
    checkNullExpr("z(0)");
    checkNullExpr("z('')");
    checkNullExpr("z(0.0)");
    checkNullExpr("z(0.00000000000001)");
    checkNullExpr("z(false)");

    // Non-Null
    checkIntExpr("z(1)", 1);
    checkIntExpr("z(999)", 999);
    checkStringExpr("z('a')", "a");
    checkFloatExpr("z(0.1)", 0.1);
    checkBoolExpr("z(true)", true);

    // Null, using Zap
    checkNullExpr("zap(0)");
    checkNullExpr("zap('')");
    checkNullExpr("zap(0.0)");
    checkNullExpr("zap(0.00000000000001)");
    checkNullExpr("zap(false)");

    // Non-Null, using Zap
    checkIntExpr("zap(1)", 1);
    checkIntExpr("zap(999)", 999);
    checkStringExpr("zap('a')", "a");
    checkFloatExpr("zap(0.1)", 0.1);
    checkBoolExpr("zap(true)", true);
}

/** Test Len function.
    Instruction: ulen */
void
IntBuiltinTestSuite::testLen()
{
    checkIntExpr("len('')", 0);
    checkIntExpr("len('a')", 1);
    checkIntExpr("len('foobar')", 6);
    checkNullExpr("len(z(0))");

    // Those yield 1 and 2, respectively, in PCC1. PCC2 does not implicitly stringify.
    checkExecFailure("len(2)");
    checkExecFailure("len(12)");
}

/** Test String/String$ function.
    Instruction: bstrmult */
void
IntBuiltinTestSuite::testStrMult()
{
    // String
    checkStringExpr("string(10)", "          ");
    checkStringExpr("string(0)", "");
    checkStringExpr("string(-1)", "");

    checkStringExpr("string(10, '')", "");
    checkStringExpr("string(10, 'a')", "aaaaaaaaaa");
    checkStringExpr("string(5, 'ab')", "ababababab");

    checkNullExpr("string(z(0), 10)");
    checkNullExpr("string(10, z(0))");
    checkNullExpr("string(z(0), z(0))");
    checkNullExpr("string('a', z(0))");
    checkNullExpr("string(z(0))");

    checkExecFailure("string('a', 10)");
    checkExecFailure("string('a', 'b')");
    checkExecFailure("string(1,2)");

    // Same thing using String$
    checkStringExpr("string$(10)", "          ");
    checkStringExpr("string$(0)", "");
    checkStringExpr("string$(-1)", "");

    checkStringExpr("string$(10, '')", "");
    checkStringExpr("string$(10, 'a')", "aaaaaaaaaa");
    checkStringExpr("string$(5, 'ab')", "ababababab");

    checkNullExpr("string$(z(0), 10)");
    checkNullExpr("string$(10, z(0))");
    checkNullExpr("string$(z(0), z(0))");
    checkNullExpr("string$('a', z(0))");
    checkNullExpr("string$(z(0))");

    checkExecFailure("string$('a', 10)");
    checkExecFailure("string$('a', 'b')");
    checkExecFailure("string$(1,2)");
}

/** Test StrCase function.
    Instructions affected by StrCase have already been tested elsewhere,
    so this tests that StrCase doesn't affect too much. */
void
IntBuiltinTestSuite::testStrCase()
{
    checkIntExpr("strcase(1+1)", 2);

    checkIntExpr("strcase(instr('foo','O')) + instr('foo','O')", 2);
    checkIntExpr("instr('foo','O') + strcase(instr('foo','O'))", 2);
    checkIntExpr("strcase(instr('foo','O')) + strcase(instr('foo','O'))", 0);
    checkIntExpr("instr('foo','O') + instr('foo','O')", 4);
    checkIntExpr("strcase(instr('foo','O') + instr('foo','O'))", 0);
    checkIntExpr("instr(strcase('foo'),'O')", 2);
    checkIntExpr("instr('foo',strcase('O'))", 2);

    checkIntExpr("if(strcase(1 or 2),3,4)", 3);

    checkIntExpr("if(strcase(instr('foo','O') or instr('foo','O')),3,4)", 4);
    checkIntExpr("if(strcase(instr('foo','O') or instr('foo','o')),3,4)", 3);
    checkIntExpr("if(strcase(instr('foo','o') or instr('foo','O')),3,4)", 3);

    checkIntExpr("if('a' = 'A', 3, 4)", 3);
    checkIntExpr("if('a' <> 'A', 3, 4)", 4);
}

/** Test atom functions: Atom, AtomStr.
    Instructions: uatom, uatomstr */
void
IntBuiltinTestSuite::testAtom()
{
    checkIntExpr("atom('')", 0);
    checkStringExpr("atomstr(0)", "");

    checkNullExpr("atom(z(0))");
    checkNullExpr("atomstr(z(0))");

    checkStringExpr("atomstr(atom(1))", "1");
    checkStringExpr("atomstr(atom('haha'))", "haha");
}

/** Test Eval function.
    Instructions: sevalx */
void
IntBuiltinTestSuite::testEval()
{
    checkIntExpr("eval(1)", 1);
    checkIntExpr("eval('1')", 1);
    checkIntExpr("eval('1+1')", 2);
    checkIntExpr("eval('1;2;3')", 3);
    checkNullExpr("eval(z(0))");
    checkNullExpr("eval('z(0)')");
    checkStringExpr("eval('\"foo\"')", "foo");
}
