/**
  *  \file util/plugin.cc
  *  \brief Plugin
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "util/plugin.h"
#include "util/confpars.h"
#include "util/console.h"
#include "util/nls.h"
#include "cpluslib/format.h"

static void
addVersion(Plugin::FeatureSet& out, string_t comp)
{
    string_t::size_type n = comp.find_first_of(" \t");
    string_t version;
    if (n != string_t::npos) {
        version = strTrim(string_t(comp, n, comp.size()-n));
        comp.erase(n);
    }
    comp = strUCase(comp);
    Plugin::FeatureSet::iterator it = out.find(comp);
    if (it != out.end()) {
        // Someone did "required = foo 1.0, foo 2.0". This is stupid.
        // Turn it into "required = foo 2.0".
        if (compareVersions(it->second, version)) {
            it->second = version;
        }
    } else {
        out.insert(std::make_pair(comp, version));
    }
}

static void
addVersions(Plugin::FeatureSet& out, string_t in)
{
    string_t::size_type n = 0, p;
    while ((p = in.find(',', n)) != string_t::npos) {
        addVersion(out, strTrim(string_t(in, n, p-n)));
        n = p+1;
    }
    addVersion(out, strTrim(string_t(in, n, in.size()-n)));
}

static void
writeVersion(TextFile& tf, const Plugin::FeatureSet::value_type& v)
{
    tf.write(v.first);
    if (v.second.empty()) {
        tf.writeLine();
    } else {
        tf.write(" ");
        tf.writeLine(v.second);
    }
}

/********************************* Plugin ********************************/

Plugin::Plugin(string_t id)
    : id(id),
      name(id),
      description(),
      baseDir(),
      defFileName(),
      provides(),
      requires(),
      items(),
      loaded(false)
{
    provides.insert(std::make_pair(id, string_t()));
}

Plugin::~Plugin()
{ }

void
Plugin::initFromPluginFile(string_t baseDir, string_t defFileName, Stream& file)
{
    class Parser : public ConfigParser {
     public:
        Parser(Plugin& self)
            : self(self)
            { }
        void assign(int /*line_id*/, string_t name, string_t value, string_t /*line*/)
            {
                if (strCaseCmp(name, "name") == 0) {
                    self.name = value;
                } else if (strCaseCmp(name, "description") == 0) {
                    if (!self.description.empty()) {
                        self.description += "\n";
                    }
                    self.description += value;
                } else if (strCaseCmp(name, "provides") == 0) {
                    addVersions(self.provides, value);
                } else if (strCaseCmp(name, "requires") == 0) {
                    addVersions(self.requires, value);
                } else if (strCaseCmp(name, "scriptfile") == 0) {
                    self.addItem(ScriptFile, value);
                } else if (strCaseCmp(name, "resourcefile") == 0) {
                    self.addItem(ResourceFile, value);
                } else if (strCaseCmp(name, "helpfile") == 0) {
                    self.addItem(HelpFile, value);
                } else if (name.size() >= 4 && strCaseCmp(name.substr(name.size()-4), "file") == 0) {
                    self.addItem(PlainFile, value);
                } else if (strCaseCmp(name, "exec") == 0) {
                    self.addItem(Command, value);
                } else {
                    // Ignore unknown key
                }
            }

        void error(int line_id, string_t msg)
            {
                console.write(LOG_WARN, format(_("%s:%d: %s -- line has been ignored"), getFileName(), line_id, msg));
            }
     private:
        Plugin& self;
    };

    this->baseDir = baseDir;
    this->defFileName = defFileName;
    Parser p(*this);
    p.configure("plugin", true);
    p.parseFile(file);
}

void
Plugin::initFromResourceFile(string_t baseDir, string_t resFileName)
{
    this->baseDir = baseDir;

    // Need not set this->defFileName which is only needed when we start with a .c2p file
    // this->defFileName = "";

    this->name = strLCase(resFileName);
    this->description = _("Resource file (artwork)");
    items.push_back(Item(ResourceFile, resFileName));
}

void
Plugin::initFromScriptFile(string_t baseDir, string_t scriptFileName, Stream& file)
{
    this->baseDir = baseDir;

    // Need not set this->defFileName which is only needed when we start with a .c2p file
    // this->defFileName = "";
    this->name = strLCase(scriptFileName);
    this->description = "";
    items.push_back(Item(ScriptFile, scriptFileName));

    // Parse the script and attempt to extract some information.
    TextFile tf(file);
    string_t line;
    enum { FindName, FindBlank, FindDescription, Stop } state = FindName;
    while (state != Stop && tf.getLine(line)) {
        // Stop when the top comment ends
        if (line.empty() || line[0] != '%') {
            break;
        }

        // Trim leading space
        size_t n = 1;
        while (line.size() > n && line[n] == ' ') {
            ++n;
        }
        line.erase(0, n);

        switch (state) {
         case FindName:
            if (!line.empty()) {
                this->name = line;
                state = FindBlank;
            }
            break;

         case FindBlank:
            if (!line.empty()) {
                // We're expecting the blank line after the description,
                // but got a nonblank line. Make it all the description.
                this->description = this->name;
                this->description += " ";
                this->description += line;
            }
            state = FindDescription;
            break;

         case FindDescription:
            if (!line.empty()) {
                if (!this->description.empty()) {
                    this->description += " ";
                }
                this->description += line;
            } else {
                state = Stop;
            }
            break;

         case Stop:
            break;
        }
    }

    // Trim description to two sentences.
    string_t::size_type n = description.find(". ");
    if (n != string_t::npos) {
        n = description.find(". ", n+1);
    }
    if (n != string_t::npos) {
        description.erase(n+1);
    }
    if (description.empty()) {
        description = _("Script file");
    }
}

void
Plugin::savePluginFile(Stream& file) const
{
    TextFile tf(file);
    tf.writeLine("# Auto-generated plugin definition file");

    tf.write("Name = ");
    tf.writeLine(name);

    string_t::size_type n = 0, p;
    while ((p = description.find('\n', n)) != string_t::npos) {
        tf.write("Description = ");
        tf.writeLine(description.substr(n, p-n));
        n = p+1;
    }
    tf.write("Description = ");
    tf.writeLine(description.substr(n));

    for (FeatureSet::const_iterator i = provides.begin(); i != provides.end(); ++i) {
        if (i->first != id) {
            tf.write("Provides = ");
            writeVersion(tf, *i);
        }
    }
    for (FeatureSet::const_iterator i = requires.begin(); i != requires.end(); ++i) {
        tf.write("Requires = ");
        writeVersion(tf, *i);
    }
    for (ItemList::const_iterator i = items.begin(); i != items.end(); ++i) {
        switch (i->type) {
         case PlainFile:    tf.write("File = ");         break;
         case ScriptFile:   tf.write("ScriptFile = ");   break;
         case ResourceFile: tf.write("ResourceFile = "); break;
         case HelpFile:     tf.write("HelpFile = ");     break;
         case Command:      tf.write("Exec = ");         break;
        }
        tf.writeLine(i->name);
    }
    tf.sync();
}

void
Plugin::setBaseDirectory(const string_t& baseDir)
{
    this->baseDir = baseDir;
}

void
Plugin::addItem(ItemType type, const string_t& name)
{
    items.push_back(Item(type, name));
}

/** Check whether this plugin provides a feature. */
bool
Plugin::isProvided(const string_t& feature) const
{
    return provides.find(feature) != provides.end();
}

/** Check for installation conflict. */
bool
Plugin::isConflict(const Plugin& other) const
{
    for (FeatureSet::const_iterator pi = provides.begin(), pe = provides.end(); pi != pe; ++pi) {
        if (other.provides.find(pi->first) != other.provides.end()) {
            // other plugin provides same feature as we -> reject
            return true;
        }
    }
    return false;
}

/** Check whether this plugin is a drop-in update for another one.
    This means it must have the same or fewer preconditions,
    and provide the same or better features. */
bool
Plugin::isUpdateFor(const Plugin& other) const
{
    // All of our preconditions must already be required by other
    for (FeatureSet::const_iterator pi = requires.begin(), pe = requires.end(); pi != pe; ++pi) {
        FeatureSet::const_iterator oi = other.requires.find(pi->first);
        if (oi == other.requires.end() || compareVersions(oi->second, pi->second)) {
            return false;
        }
    }

    // All of other's features must be provided by us
    for (FeatureSet::const_iterator pi = other.provides.begin(), pe = other.provides.end(); pi != pe; ++pi) {
        FeatureSet::const_iterator oi = provides.find(pi->first);
        if (oi == provides.end() || compareVersions(oi->second, pi->second)) {
            return false;
        }
    }

    return true;
}

/** Check whether this plugin depends on another one. */
bool
Plugin::isDependingOn(const Plugin& other) const
{
    // Since only one plugin can provide a feature, this condition is satisfied
    // if one of our required features is provided by %other. There cannot be
    // a different plugin that provides the same feature.
    for (FeatureSet::const_iterator pi = requires.begin(), pe = requires.end(); pi != pe; ++pi) {
        if (other.provides.find(pi->first) != other.provides.end()) {
            return true;
        }
    }
    return false;
}

/** Check whether the given feature set satisfies our requirements. */
bool
Plugin::isSatisfied(const FeatureSet& fset) const
{
    for (FeatureSet::const_iterator pi = requires.begin(), pe = requires.end(); pi != pe; ++pi) {
        FeatureSet::const_iterator po = fset.find(pi->first);
        if (po == fset.end() || compareVersions(po->second, pi->second)) {
            // Feature not available or wrong version
            return false;
        }
    }
    return true;
}

void
Plugin::enumMissingFeatures(const FeatureSet& have, FeatureSet& missing) const
{
    for (FeatureSet::const_iterator pi = requires.begin(), pe = requires.end(); pi != pe; ++pi) {
        FeatureSet::const_iterator po = have.find(pi->first);
        if (po == have.end() || compareVersions(po->second, pi->second)) {
            missing.insert(*pi);
        }
    }
}

/** Add all features we provide to the given feature set. */
void
Plugin::addProvidedFeaturesTo(FeatureSet& fset)
{
    for (FeatureSet::const_iterator pi = provides.begin(), pe = provides.end(); pi != pe; ++pi) {
        // No version check required; we are the only one who provides this feature,
        // so there cannot be a previous version to upgrade.
        fset.insert(*pi);
    }
}

static bool
eatVersion(const string_t& str, string_t::size_type& pos, uint32_t& version)
{
    bool ok = false;
    version = 0;
    while (pos < str.size() && str[pos] >= '0' && str[pos] <= '9') {
        version = 10*version + (str[pos] - '0');
        ++pos;
        ok = true;
    }
    if (ok && pos < str.size() && str[pos] == '.') {
        ++pos;
    }
    return ok;
}

/** Compare versions.
    \return true iff a is older than b */
bool
compareVersions(const string_t& a, const string_t& b)
{
    string_t::size_type apos = 0, bpos = 0;
    while (1) {
        uint32_t avers, bvers;
        bool aok = eatVersion(a, apos, avers);
        bool bok = eatVersion(b, bpos, bvers);
        if (aok) {
            if (bok) {
                // Two versions.
                if (avers != bvers) {
                    return avers < bvers;
                }
            } else {
                // One version, e.g. "1.1" vs. "1.foo".
                // Treat second one older.
                return false;
            }
        } else {
            if (bok) {
                // One version, e.g. "1.foo" vs. "1.1"
                // Treat first one as older.
                return true;
            } else {
                // No version, e.g. "1.a" vs. "1.b".
                // String compare.
                return a.substr(apos) < b.substr(bpos);
            }
        }
    }
}
